# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from openapi_client.configuration import Configuration


class ResponsePipelineStepJupyterVisualizationSpec(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'env_vars': 'OrmEnvVars',
        'kernel_cluster': 'ResponseKernelCluster',
        'kernel_cluster_select_policies': 'OrmKernelClusterSelectPolicies',
        'kernel_image': 'ResponseKernelImage',
        'kernel_resource_spec': 'ResponseKernelResourceSpec',
        'start_command': 'str',
        'variable_specs': 'list[PipelinePipelineSingleVariableSpec]',
        'volume_tree': 'VolumetreenodeVolumeTreeNode'
    }

    attribute_map = {
        'env_vars': 'env_vars',
        'kernel_cluster': 'kernel_cluster',
        'kernel_cluster_select_policies': 'kernel_cluster_select_policies',
        'kernel_image': 'kernel_image',
        'kernel_resource_spec': 'kernel_resource_spec',
        'start_command': 'start_command',
        'variable_specs': 'variable_specs',
        'volume_tree': 'volume_tree'
    }

    def __init__(self, env_vars=None, kernel_cluster=None, kernel_cluster_select_policies=None, kernel_image=None, kernel_resource_spec=None, start_command=None, variable_specs=None, volume_tree=None, local_vars_configuration=None):  # noqa: E501
        """ResponsePipelineStepJupyterVisualizationSpec - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._env_vars = None
        self._kernel_cluster = None
        self._kernel_cluster_select_policies = None
        self._kernel_image = None
        self._kernel_resource_spec = None
        self._start_command = None
        self._variable_specs = None
        self._volume_tree = None
        self.discriminator = None

        if env_vars is not None:
            self.env_vars = env_vars
        if kernel_cluster is not None:
            self.kernel_cluster = kernel_cluster
        if kernel_cluster_select_policies is not None:
            self.kernel_cluster_select_policies = kernel_cluster_select_policies
        if kernel_image is not None:
            self.kernel_image = kernel_image
        if kernel_resource_spec is not None:
            self.kernel_resource_spec = kernel_resource_spec
        self.start_command = start_command
        if variable_specs is not None:
            self.variable_specs = variable_specs
        if volume_tree is not None:
            self.volume_tree = volume_tree

    @property
    def env_vars(self):
        """Gets the env_vars of this ResponsePipelineStepJupyterVisualizationSpec.  # noqa: E501


        :return: The env_vars of this ResponsePipelineStepJupyterVisualizationSpec.  # noqa: E501
        :rtype: OrmEnvVars
        """
        return self._env_vars

    @env_vars.setter
    def env_vars(self, env_vars):
        """Sets the env_vars of this ResponsePipelineStepJupyterVisualizationSpec.


        :param env_vars: The env_vars of this ResponsePipelineStepJupyterVisualizationSpec.  # noqa: E501
        :type env_vars: OrmEnvVars
        """

        self._env_vars = env_vars

    @property
    def kernel_cluster(self):
        """Gets the kernel_cluster of this ResponsePipelineStepJupyterVisualizationSpec.  # noqa: E501


        :return: The kernel_cluster of this ResponsePipelineStepJupyterVisualizationSpec.  # noqa: E501
        :rtype: ResponseKernelCluster
        """
        return self._kernel_cluster

    @kernel_cluster.setter
    def kernel_cluster(self, kernel_cluster):
        """Sets the kernel_cluster of this ResponsePipelineStepJupyterVisualizationSpec.


        :param kernel_cluster: The kernel_cluster of this ResponsePipelineStepJupyterVisualizationSpec.  # noqa: E501
        :type kernel_cluster: ResponseKernelCluster
        """

        self._kernel_cluster = kernel_cluster

    @property
    def kernel_cluster_select_policies(self):
        """Gets the kernel_cluster_select_policies of this ResponsePipelineStepJupyterVisualizationSpec.  # noqa: E501


        :return: The kernel_cluster_select_policies of this ResponsePipelineStepJupyterVisualizationSpec.  # noqa: E501
        :rtype: OrmKernelClusterSelectPolicies
        """
        return self._kernel_cluster_select_policies

    @kernel_cluster_select_policies.setter
    def kernel_cluster_select_policies(self, kernel_cluster_select_policies):
        """Sets the kernel_cluster_select_policies of this ResponsePipelineStepJupyterVisualizationSpec.


        :param kernel_cluster_select_policies: The kernel_cluster_select_policies of this ResponsePipelineStepJupyterVisualizationSpec.  # noqa: E501
        :type kernel_cluster_select_policies: OrmKernelClusterSelectPolicies
        """

        self._kernel_cluster_select_policies = kernel_cluster_select_policies

    @property
    def kernel_image(self):
        """Gets the kernel_image of this ResponsePipelineStepJupyterVisualizationSpec.  # noqa: E501


        :return: The kernel_image of this ResponsePipelineStepJupyterVisualizationSpec.  # noqa: E501
        :rtype: ResponseKernelImage
        """
        return self._kernel_image

    @kernel_image.setter
    def kernel_image(self, kernel_image):
        """Sets the kernel_image of this ResponsePipelineStepJupyterVisualizationSpec.


        :param kernel_image: The kernel_image of this ResponsePipelineStepJupyterVisualizationSpec.  # noqa: E501
        :type kernel_image: ResponseKernelImage
        """

        self._kernel_image = kernel_image

    @property
    def kernel_resource_spec(self):
        """Gets the kernel_resource_spec of this ResponsePipelineStepJupyterVisualizationSpec.  # noqa: E501


        :return: The kernel_resource_spec of this ResponsePipelineStepJupyterVisualizationSpec.  # noqa: E501
        :rtype: ResponseKernelResourceSpec
        """
        return self._kernel_resource_spec

    @kernel_resource_spec.setter
    def kernel_resource_spec(self, kernel_resource_spec):
        """Sets the kernel_resource_spec of this ResponsePipelineStepJupyterVisualizationSpec.


        :param kernel_resource_spec: The kernel_resource_spec of this ResponsePipelineStepJupyterVisualizationSpec.  # noqa: E501
        :type kernel_resource_spec: ResponseKernelResourceSpec
        """

        self._kernel_resource_spec = kernel_resource_spec

    @property
    def start_command(self):
        """Gets the start_command of this ResponsePipelineStepJupyterVisualizationSpec.  # noqa: E501


        :return: The start_command of this ResponsePipelineStepJupyterVisualizationSpec.  # noqa: E501
        :rtype: str
        """
        return self._start_command

    @start_command.setter
    def start_command(self, start_command):
        """Sets the start_command of this ResponsePipelineStepJupyterVisualizationSpec.


        :param start_command: The start_command of this ResponsePipelineStepJupyterVisualizationSpec.  # noqa: E501
        :type start_command: str
        """
        if self.local_vars_configuration.client_side_validation and start_command is None:  # noqa: E501
            raise ValueError("Invalid value for `start_command`, must not be `None`")  # noqa: E501

        self._start_command = start_command

    @property
    def variable_specs(self):
        """Gets the variable_specs of this ResponsePipelineStepJupyterVisualizationSpec.  # noqa: E501


        :return: The variable_specs of this ResponsePipelineStepJupyterVisualizationSpec.  # noqa: E501
        :rtype: list[PipelinePipelineSingleVariableSpec]
        """
        return self._variable_specs

    @variable_specs.setter
    def variable_specs(self, variable_specs):
        """Sets the variable_specs of this ResponsePipelineStepJupyterVisualizationSpec.


        :param variable_specs: The variable_specs of this ResponsePipelineStepJupyterVisualizationSpec.  # noqa: E501
        :type variable_specs: list[PipelinePipelineSingleVariableSpec]
        """

        self._variable_specs = variable_specs

    @property
    def volume_tree(self):
        """Gets the volume_tree of this ResponsePipelineStepJupyterVisualizationSpec.  # noqa: E501


        :return: The volume_tree of this ResponsePipelineStepJupyterVisualizationSpec.  # noqa: E501
        :rtype: VolumetreenodeVolumeTreeNode
        """
        return self._volume_tree

    @volume_tree.setter
    def volume_tree(self, volume_tree):
        """Sets the volume_tree of this ResponsePipelineStepJupyterVisualizationSpec.


        :param volume_tree: The volume_tree of this ResponsePipelineStepJupyterVisualizationSpec.  # noqa: E501
        :type volume_tree: VolumetreenodeVolumeTreeNode
        """

        self._volume_tree = volume_tree

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ResponsePipelineStepJupyterVisualizationSpec):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ResponsePipelineStepJupyterVisualizationSpec):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from openapi_client.configuration import Configuration


class ResponseKernelClusterInfo(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'available_gpus': 'list[str]',
        'config': 'OrmKernelClusterConfig',
        'created_dt': 'datetime',
        'current_system_metrics': 'list[InfluxdbCurrentSystemMetric]',
        'docs_link': 'str',
        'id': 'int',
        'is_cluster_admin': 'bool',
        'is_custom_resource_spec_allowed': 'bool',
        'is_registered': 'bool',
        'is_savvihub_managed': 'bool',
        'is_vessl_managed': 'bool',
        'issues': 'list[str]',
        'kubernetes_ingress_endpoint': 'str',
        'kubernetes_namespace': 'str',
        'kubernetes_service_type': 'str',
        'latest_cluster_metrics': 'PrometheusqueryClusterMetricSample',
        'message': 'str',
        'metric_source': 'str',
        'name': 'str',
        'nodes_summary': 'dict[str, int]',
        'provider': 'str',
        'region': 'str',
        'status': 'str',
        'updated_dt': 'datetime',
        'workloads_summary': 'OrmWorkloadsSummary'
    }

    attribute_map = {
        'available_gpus': 'available_gpus',
        'config': 'config',
        'created_dt': 'created_dt',
        'current_system_metrics': 'current_system_metrics',
        'docs_link': 'docs_link',
        'id': 'id',
        'is_cluster_admin': 'is_cluster_admin',
        'is_custom_resource_spec_allowed': 'is_custom_resource_spec_allowed',
        'is_registered': 'is_registered',
        'is_savvihub_managed': 'is_savvihub_managed',
        'is_vessl_managed': 'is_vessl_managed',
        'issues': 'issues',
        'kubernetes_ingress_endpoint': 'kubernetes_ingress_endpoint',
        'kubernetes_namespace': 'kubernetes_namespace',
        'kubernetes_service_type': 'kubernetes_service_type',
        'latest_cluster_metrics': 'latest_cluster_metrics',
        'message': 'message',
        'metric_source': 'metric_source',
        'name': 'name',
        'nodes_summary': 'nodes_summary',
        'provider': 'provider',
        'region': 'region',
        'status': 'status',
        'updated_dt': 'updated_dt',
        'workloads_summary': 'workloads_summary'
    }

    def __init__(self, available_gpus=None, config=None, created_dt=None, current_system_metrics=None, docs_link=None, id=None, is_cluster_admin=None, is_custom_resource_spec_allowed=None, is_registered=None, is_savvihub_managed=None, is_vessl_managed=None, issues=None, kubernetes_ingress_endpoint=None, kubernetes_namespace=None, kubernetes_service_type=None, latest_cluster_metrics=None, message=None, metric_source=None, name=None, nodes_summary=None, provider=None, region=None, status=None, updated_dt=None, workloads_summary=None, local_vars_configuration=None):  # noqa: E501
        """ResponseKernelClusterInfo - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._available_gpus = None
        self._config = None
        self._created_dt = None
        self._current_system_metrics = None
        self._docs_link = None
        self._id = None
        self._is_cluster_admin = None
        self._is_custom_resource_spec_allowed = None
        self._is_registered = None
        self._is_savvihub_managed = None
        self._is_vessl_managed = None
        self._issues = None
        self._kubernetes_ingress_endpoint = None
        self._kubernetes_namespace = None
        self._kubernetes_service_type = None
        self._latest_cluster_metrics = None
        self._message = None
        self._metric_source = None
        self._name = None
        self._nodes_summary = None
        self._provider = None
        self._region = None
        self._status = None
        self._updated_dt = None
        self._workloads_summary = None
        self.discriminator = None

        if available_gpus is not None:
            self.available_gpus = available_gpus
        if config is not None:
            self.config = config
        self.created_dt = created_dt
        if current_system_metrics is not None:
            self.current_system_metrics = current_system_metrics
        self.docs_link = docs_link
        self.id = id
        if is_cluster_admin is not None:
            self.is_cluster_admin = is_cluster_admin
        self.is_custom_resource_spec_allowed = is_custom_resource_spec_allowed
        if is_registered is not None:
            self.is_registered = is_registered
        self.is_savvihub_managed = is_savvihub_managed
        self.is_vessl_managed = is_vessl_managed
        if issues is not None:
            self.issues = issues
        self.kubernetes_ingress_endpoint = kubernetes_ingress_endpoint
        self.kubernetes_namespace = kubernetes_namespace
        self.kubernetes_service_type = kubernetes_service_type
        if latest_cluster_metrics is not None:
            self.latest_cluster_metrics = latest_cluster_metrics
        self.message = message
        self.metric_source = metric_source
        self.name = name
        if nodes_summary is not None:
            self.nodes_summary = nodes_summary
        self.provider = provider
        self.region = region
        self.status = status
        self.updated_dt = updated_dt
        if workloads_summary is not None:
            self.workloads_summary = workloads_summary

    @property
    def available_gpus(self):
        """Gets the available_gpus of this ResponseKernelClusterInfo.  # noqa: E501


        :return: The available_gpus of this ResponseKernelClusterInfo.  # noqa: E501
        :rtype: list[str]
        """
        return self._available_gpus

    @available_gpus.setter
    def available_gpus(self, available_gpus):
        """Sets the available_gpus of this ResponseKernelClusterInfo.


        :param available_gpus: The available_gpus of this ResponseKernelClusterInfo.  # noqa: E501
        :type available_gpus: list[str]
        """

        self._available_gpus = available_gpus

    @property
    def config(self):
        """Gets the config of this ResponseKernelClusterInfo.  # noqa: E501


        :return: The config of this ResponseKernelClusterInfo.  # noqa: E501
        :rtype: OrmKernelClusterConfig
        """
        return self._config

    @config.setter
    def config(self, config):
        """Sets the config of this ResponseKernelClusterInfo.


        :param config: The config of this ResponseKernelClusterInfo.  # noqa: E501
        :type config: OrmKernelClusterConfig
        """

        self._config = config

    @property
    def created_dt(self):
        """Gets the created_dt of this ResponseKernelClusterInfo.  # noqa: E501


        :return: The created_dt of this ResponseKernelClusterInfo.  # noqa: E501
        :rtype: datetime
        """
        return self._created_dt

    @created_dt.setter
    def created_dt(self, created_dt):
        """Sets the created_dt of this ResponseKernelClusterInfo.


        :param created_dt: The created_dt of this ResponseKernelClusterInfo.  # noqa: E501
        :type created_dt: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_dt is None:  # noqa: E501
            raise ValueError("Invalid value for `created_dt`, must not be `None`")  # noqa: E501

        self._created_dt = created_dt

    @property
    def current_system_metrics(self):
        """Gets the current_system_metrics of this ResponseKernelClusterInfo.  # noqa: E501


        :return: The current_system_metrics of this ResponseKernelClusterInfo.  # noqa: E501
        :rtype: list[InfluxdbCurrentSystemMetric]
        """
        return self._current_system_metrics

    @current_system_metrics.setter
    def current_system_metrics(self, current_system_metrics):
        """Sets the current_system_metrics of this ResponseKernelClusterInfo.


        :param current_system_metrics: The current_system_metrics of this ResponseKernelClusterInfo.  # noqa: E501
        :type current_system_metrics: list[InfluxdbCurrentSystemMetric]
        """

        self._current_system_metrics = current_system_metrics

    @property
    def docs_link(self):
        """Gets the docs_link of this ResponseKernelClusterInfo.  # noqa: E501


        :return: The docs_link of this ResponseKernelClusterInfo.  # noqa: E501
        :rtype: str
        """
        return self._docs_link

    @docs_link.setter
    def docs_link(self, docs_link):
        """Sets the docs_link of this ResponseKernelClusterInfo.


        :param docs_link: The docs_link of this ResponseKernelClusterInfo.  # noqa: E501
        :type docs_link: str
        """

        self._docs_link = docs_link

    @property
    def id(self):
        """Gets the id of this ResponseKernelClusterInfo.  # noqa: E501


        :return: The id of this ResponseKernelClusterInfo.  # noqa: E501
        :rtype: int
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this ResponseKernelClusterInfo.


        :param id: The id of this ResponseKernelClusterInfo.  # noqa: E501
        :type id: int
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def is_cluster_admin(self):
        """Gets the is_cluster_admin of this ResponseKernelClusterInfo.  # noqa: E501


        :return: The is_cluster_admin of this ResponseKernelClusterInfo.  # noqa: E501
        :rtype: bool
        """
        return self._is_cluster_admin

    @is_cluster_admin.setter
    def is_cluster_admin(self, is_cluster_admin):
        """Sets the is_cluster_admin of this ResponseKernelClusterInfo.


        :param is_cluster_admin: The is_cluster_admin of this ResponseKernelClusterInfo.  # noqa: E501
        :type is_cluster_admin: bool
        """

        self._is_cluster_admin = is_cluster_admin

    @property
    def is_custom_resource_spec_allowed(self):
        """Gets the is_custom_resource_spec_allowed of this ResponseKernelClusterInfo.  # noqa: E501


        :return: The is_custom_resource_spec_allowed of this ResponseKernelClusterInfo.  # noqa: E501
        :rtype: bool
        """
        return self._is_custom_resource_spec_allowed

    @is_custom_resource_spec_allowed.setter
    def is_custom_resource_spec_allowed(self, is_custom_resource_spec_allowed):
        """Sets the is_custom_resource_spec_allowed of this ResponseKernelClusterInfo.


        :param is_custom_resource_spec_allowed: The is_custom_resource_spec_allowed of this ResponseKernelClusterInfo.  # noqa: E501
        :type is_custom_resource_spec_allowed: bool
        """
        if self.local_vars_configuration.client_side_validation and is_custom_resource_spec_allowed is None:  # noqa: E501
            raise ValueError("Invalid value for `is_custom_resource_spec_allowed`, must not be `None`")  # noqa: E501

        self._is_custom_resource_spec_allowed = is_custom_resource_spec_allowed

    @property
    def is_registered(self):
        """Gets the is_registered of this ResponseKernelClusterInfo.  # noqa: E501


        :return: The is_registered of this ResponseKernelClusterInfo.  # noqa: E501
        :rtype: bool
        """
        return self._is_registered

    @is_registered.setter
    def is_registered(self, is_registered):
        """Sets the is_registered of this ResponseKernelClusterInfo.


        :param is_registered: The is_registered of this ResponseKernelClusterInfo.  # noqa: E501
        :type is_registered: bool
        """

        self._is_registered = is_registered

    @property
    def is_savvihub_managed(self):
        """Gets the is_savvihub_managed of this ResponseKernelClusterInfo.  # noqa: E501


        :return: The is_savvihub_managed of this ResponseKernelClusterInfo.  # noqa: E501
        :rtype: bool
        """
        return self._is_savvihub_managed

    @is_savvihub_managed.setter
    def is_savvihub_managed(self, is_savvihub_managed):
        """Sets the is_savvihub_managed of this ResponseKernelClusterInfo.


        :param is_savvihub_managed: The is_savvihub_managed of this ResponseKernelClusterInfo.  # noqa: E501
        :type is_savvihub_managed: bool
        """
        if self.local_vars_configuration.client_side_validation and is_savvihub_managed is None:  # noqa: E501
            raise ValueError("Invalid value for `is_savvihub_managed`, must not be `None`")  # noqa: E501

        self._is_savvihub_managed = is_savvihub_managed

    @property
    def is_vessl_managed(self):
        """Gets the is_vessl_managed of this ResponseKernelClusterInfo.  # noqa: E501


        :return: The is_vessl_managed of this ResponseKernelClusterInfo.  # noqa: E501
        :rtype: bool
        """
        return self._is_vessl_managed

    @is_vessl_managed.setter
    def is_vessl_managed(self, is_vessl_managed):
        """Sets the is_vessl_managed of this ResponseKernelClusterInfo.


        :param is_vessl_managed: The is_vessl_managed of this ResponseKernelClusterInfo.  # noqa: E501
        :type is_vessl_managed: bool
        """
        if self.local_vars_configuration.client_side_validation and is_vessl_managed is None:  # noqa: E501
            raise ValueError("Invalid value for `is_vessl_managed`, must not be `None`")  # noqa: E501

        self._is_vessl_managed = is_vessl_managed

    @property
    def issues(self):
        """Gets the issues of this ResponseKernelClusterInfo.  # noqa: E501


        :return: The issues of this ResponseKernelClusterInfo.  # noqa: E501
        :rtype: list[str]
        """
        return self._issues

    @issues.setter
    def issues(self, issues):
        """Sets the issues of this ResponseKernelClusterInfo.


        :param issues: The issues of this ResponseKernelClusterInfo.  # noqa: E501
        :type issues: list[str]
        """

        self._issues = issues

    @property
    def kubernetes_ingress_endpoint(self):
        """Gets the kubernetes_ingress_endpoint of this ResponseKernelClusterInfo.  # noqa: E501


        :return: The kubernetes_ingress_endpoint of this ResponseKernelClusterInfo.  # noqa: E501
        :rtype: str
        """
        return self._kubernetes_ingress_endpoint

    @kubernetes_ingress_endpoint.setter
    def kubernetes_ingress_endpoint(self, kubernetes_ingress_endpoint):
        """Sets the kubernetes_ingress_endpoint of this ResponseKernelClusterInfo.


        :param kubernetes_ingress_endpoint: The kubernetes_ingress_endpoint of this ResponseKernelClusterInfo.  # noqa: E501
        :type kubernetes_ingress_endpoint: str
        """

        self._kubernetes_ingress_endpoint = kubernetes_ingress_endpoint

    @property
    def kubernetes_namespace(self):
        """Gets the kubernetes_namespace of this ResponseKernelClusterInfo.  # noqa: E501


        :return: The kubernetes_namespace of this ResponseKernelClusterInfo.  # noqa: E501
        :rtype: str
        """
        return self._kubernetes_namespace

    @kubernetes_namespace.setter
    def kubernetes_namespace(self, kubernetes_namespace):
        """Sets the kubernetes_namespace of this ResponseKernelClusterInfo.


        :param kubernetes_namespace: The kubernetes_namespace of this ResponseKernelClusterInfo.  # noqa: E501
        :type kubernetes_namespace: str
        """

        self._kubernetes_namespace = kubernetes_namespace

    @property
    def kubernetes_service_type(self):
        """Gets the kubernetes_service_type of this ResponseKernelClusterInfo.  # noqa: E501


        :return: The kubernetes_service_type of this ResponseKernelClusterInfo.  # noqa: E501
        :rtype: str
        """
        return self._kubernetes_service_type

    @kubernetes_service_type.setter
    def kubernetes_service_type(self, kubernetes_service_type):
        """Sets the kubernetes_service_type of this ResponseKernelClusterInfo.


        :param kubernetes_service_type: The kubernetes_service_type of this ResponseKernelClusterInfo.  # noqa: E501
        :type kubernetes_service_type: str
        """
        if self.local_vars_configuration.client_side_validation and kubernetes_service_type is None:  # noqa: E501
            raise ValueError("Invalid value for `kubernetes_service_type`, must not be `None`")  # noqa: E501

        self._kubernetes_service_type = kubernetes_service_type

    @property
    def latest_cluster_metrics(self):
        """Gets the latest_cluster_metrics of this ResponseKernelClusterInfo.  # noqa: E501


        :return: The latest_cluster_metrics of this ResponseKernelClusterInfo.  # noqa: E501
        :rtype: PrometheusqueryClusterMetricSample
        """
        return self._latest_cluster_metrics

    @latest_cluster_metrics.setter
    def latest_cluster_metrics(self, latest_cluster_metrics):
        """Sets the latest_cluster_metrics of this ResponseKernelClusterInfo.


        :param latest_cluster_metrics: The latest_cluster_metrics of this ResponseKernelClusterInfo.  # noqa: E501
        :type latest_cluster_metrics: PrometheusqueryClusterMetricSample
        """

        self._latest_cluster_metrics = latest_cluster_metrics

    @property
    def message(self):
        """Gets the message of this ResponseKernelClusterInfo.  # noqa: E501


        :return: The message of this ResponseKernelClusterInfo.  # noqa: E501
        :rtype: str
        """
        return self._message

    @message.setter
    def message(self, message):
        """Sets the message of this ResponseKernelClusterInfo.


        :param message: The message of this ResponseKernelClusterInfo.  # noqa: E501
        :type message: str
        """

        self._message = message

    @property
    def metric_source(self):
        """Gets the metric_source of this ResponseKernelClusterInfo.  # noqa: E501


        :return: The metric_source of this ResponseKernelClusterInfo.  # noqa: E501
        :rtype: str
        """
        return self._metric_source

    @metric_source.setter
    def metric_source(self, metric_source):
        """Sets the metric_source of this ResponseKernelClusterInfo.


        :param metric_source: The metric_source of this ResponseKernelClusterInfo.  # noqa: E501
        :type metric_source: str
        """
        if self.local_vars_configuration.client_side_validation and metric_source is None:  # noqa: E501
            raise ValueError("Invalid value for `metric_source`, must not be `None`")  # noqa: E501

        self._metric_source = metric_source

    @property
    def name(self):
        """Gets the name of this ResponseKernelClusterInfo.  # noqa: E501


        :return: The name of this ResponseKernelClusterInfo.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this ResponseKernelClusterInfo.


        :param name: The name of this ResponseKernelClusterInfo.  # noqa: E501
        :type name: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def nodes_summary(self):
        """Gets the nodes_summary of this ResponseKernelClusterInfo.  # noqa: E501


        :return: The nodes_summary of this ResponseKernelClusterInfo.  # noqa: E501
        :rtype: dict[str, int]
        """
        return self._nodes_summary

    @nodes_summary.setter
    def nodes_summary(self, nodes_summary):
        """Sets the nodes_summary of this ResponseKernelClusterInfo.


        :param nodes_summary: The nodes_summary of this ResponseKernelClusterInfo.  # noqa: E501
        :type nodes_summary: dict[str, int]
        """

        self._nodes_summary = nodes_summary

    @property
    def provider(self):
        """Gets the provider of this ResponseKernelClusterInfo.  # noqa: E501


        :return: The provider of this ResponseKernelClusterInfo.  # noqa: E501
        :rtype: str
        """
        return self._provider

    @provider.setter
    def provider(self, provider):
        """Sets the provider of this ResponseKernelClusterInfo.


        :param provider: The provider of this ResponseKernelClusterInfo.  # noqa: E501
        :type provider: str
        """
        if self.local_vars_configuration.client_side_validation and provider is None:  # noqa: E501
            raise ValueError("Invalid value for `provider`, must not be `None`")  # noqa: E501
        allowed_values = ["VESSL", "AWS", "GCP", "Azure", "on-premise"]  # noqa: E501
        if self.local_vars_configuration.client_side_validation and provider not in allowed_values:  # noqa: E501
            raise ValueError(
                "Invalid value for `provider` ({0}), must be one of {1}"  # noqa: E501
                .format(provider, allowed_values)
            )

        self._provider = provider

    @property
    def region(self):
        """Gets the region of this ResponseKernelClusterInfo.  # noqa: E501


        :return: The region of this ResponseKernelClusterInfo.  # noqa: E501
        :rtype: str
        """
        return self._region

    @region.setter
    def region(self, region):
        """Sets the region of this ResponseKernelClusterInfo.


        :param region: The region of this ResponseKernelClusterInfo.  # noqa: E501
        :type region: str
        """

        self._region = region

    @property
    def status(self):
        """Gets the status of this ResponseKernelClusterInfo.  # noqa: E501


        :return: The status of this ResponseKernelClusterInfo.  # noqa: E501
        :rtype: str
        """
        return self._status

    @status.setter
    def status(self, status):
        """Sets the status of this ResponseKernelClusterInfo.


        :param status: The status of this ResponseKernelClusterInfo.  # noqa: E501
        :type status: str
        """
        if self.local_vars_configuration.client_side_validation and status is None:  # noqa: E501
            raise ValueError("Invalid value for `status`, must not be `None`")  # noqa: E501
        allowed_values = ["connected", "not-connected", "permission-required", "agent-not-available", "prometheus-not-available", "modifying"]  # noqa: E501
        if self.local_vars_configuration.client_side_validation and status not in allowed_values:  # noqa: E501
            raise ValueError(
                "Invalid value for `status` ({0}), must be one of {1}"  # noqa: E501
                .format(status, allowed_values)
            )

        self._status = status

    @property
    def updated_dt(self):
        """Gets the updated_dt of this ResponseKernelClusterInfo.  # noqa: E501


        :return: The updated_dt of this ResponseKernelClusterInfo.  # noqa: E501
        :rtype: datetime
        """
        return self._updated_dt

    @updated_dt.setter
    def updated_dt(self, updated_dt):
        """Sets the updated_dt of this ResponseKernelClusterInfo.


        :param updated_dt: The updated_dt of this ResponseKernelClusterInfo.  # noqa: E501
        :type updated_dt: datetime
        """
        if self.local_vars_configuration.client_side_validation and updated_dt is None:  # noqa: E501
            raise ValueError("Invalid value for `updated_dt`, must not be `None`")  # noqa: E501

        self._updated_dt = updated_dt

    @property
    def workloads_summary(self):
        """Gets the workloads_summary of this ResponseKernelClusterInfo.  # noqa: E501


        :return: The workloads_summary of this ResponseKernelClusterInfo.  # noqa: E501
        :rtype: OrmWorkloadsSummary
        """
        return self._workloads_summary

    @workloads_summary.setter
    def workloads_summary(self, workloads_summary):
        """Sets the workloads_summary of this ResponseKernelClusterInfo.


        :param workloads_summary: The workloads_summary of this ResponseKernelClusterInfo.  # noqa: E501
        :type workloads_summary: OrmWorkloadsSummary
        """

        self._workloads_summary = workloads_summary

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ResponseKernelClusterInfo):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ResponseKernelClusterInfo):
            return True

        return self.to_dict() != other.to_dict()



from typing import List

from openapi_client import VESSLDatasetCreateAPIInput
from openapi_client.models import (
    GSDatasetCreateAPIInput,
    ResponseDatasetInfo,
    ResponseDatasetInfoDetail,
    ResponseDatasetVersionInfo,
    S3DatasetCreateAPIInput,
    StorageFile,
)
from vessl import vessl_api
from vessl.organization import _get_organization_name
from vessl.util.constant import DATASET_PATH_SCHEME_GS, DATASET_PATH_SCHEME_S3
from vessl.util.exception import InvalidDatasetError
from vessl.volume import (
    _copy_volume_file_remote_to_local,
    copy_volume_file,
    delete_volume_file,
    list_volume_files,
)


def read_dataset(dataset_name: str, **kwargs) -> ResponseDatasetInfoDetail:
    """Read a dataset in the default organization. If you want to override the
    default organization, then pass `organization_name` as `**kwargs`.

    Args:
        dataset_name(str): Dataset name.

    Example:
        ```python
        vessl.read_dataset(
           dataset_name="mnist",
        )
        ```
    """
    return vessl_api.dataset_read_api(
        dataset_name=dataset_name, organization_name=_get_organization_name(**kwargs)
    )


def read_dataset_version(
    dataset_id: int,
    dataset_version_hash: str,
    **kwargs,
) -> ResponseDatasetVersionInfo:
    """Read the specific version of dataset in the default organization. If you
    want to override the default organization, then pass `organization_name` as
    `**kwargs`.

    Args:
        dataset_id(int): Dataset id.
        dataset_version_hash(str): Dataset version hash.

    Example:
        ```python
        vessl.read_dataset_version(
            dataset_id=1,
            dataset_version_hash="hash123"
        )
        ```
    """
    return vessl_api.dataset_version_read_api(
        dataset_id=dataset_id,
        dataset_version_hash=dataset_version_hash,
        organization_name=_get_organization_name(**kwargs),
    )


def list_datasets(**kwargs) -> List[ResponseDatasetInfo]:
    """List datasets in the default organization. If you want to override the
    default organization, then pass `organization_name` as `**kwargs`.

    Example:
        ```
        vessl.list_datasets()
        ```
    """
    return vessl_api.dataset_list_api(
        organization_name=_get_organization_name(**kwargs),
    ).results


def _create_dataset_local(
    dataset_name: str,
    is_version_enabled: bool = False,
    description: str = None,
    **kwargs,
) -> ResponseDatasetInfoDetail:
    return vessl_api.v_essl_dataset_create_api(
        organization_name=_get_organization_name(**kwargs),
        vessl_dataset_create_api_input=VESSLDatasetCreateAPIInput(
            name=dataset_name,
            description=description,
            is_version_enabled=is_version_enabled,
        ),
    )


def _create_dataset_s3(
    dataset_name: str,
    is_version_enabled: bool = False,
    is_public: bool = True,
    description: str = None,
    external_path: str = None,
    aws_role_arn: str = None,
    version_path: str = None,
    **kwargs,
) -> ResponseDatasetInfoDetail:
    if is_version_enabled and (
        version_path is None or not version_path.startswith(DATASET_PATH_SCHEME_S3)
    ):
        raise InvalidDatasetError(f"Invalid version path: {version_path}")

    return vessl_api.s3_dataset_create_api(
        organization_name=_get_organization_name(**kwargs),
        s3_dataset_create_api_input=S3DatasetCreateAPIInput(
            name=dataset_name,
            description=description,
            is_version_enabled=is_version_enabled,
            s3_path=external_path,
            version_s3_path=version_path,
            is_public_bucket=is_public,
            aws_role_arn=aws_role_arn,
        ),
    )


def _create_dataset_gs(
    dataset_name: str,
    is_version_enabled: bool = False,
    is_public: bool = False,
    description: str = None,
    external_path: str = None,
    version_path: str = None,
    **kwargs,
) -> ResponseDatasetInfoDetail:
    if is_version_enabled:
        raise InvalidDatasetError("Versioning is not supported for GoogleStorage")

    return vessl_api.g_s_dataset_create_api(
        organization_name=_get_organization_name(**kwargs),
        gs_dataset_create_api_input=GSDatasetCreateAPIInput(
            name=dataset_name,
            description=description,
            is_version_enabled=is_version_enabled,
            gs_path=external_path,
            version_gs_path=version_path,
            is_public=is_public,
        ),
    )


def create_dataset(
    dataset_name: str,
    description: str = None,
    is_version_enabled: bool = False,
    is_public: bool = False,
    external_path: str = None,
    aws_role_arn: str = None,
    version_path: str = None,
    **kwargs,
) -> ResponseDatasetInfoDetail:
    """Create a dataset in the default organization. If you want to override
    the default organization, then pass `organization_name` as `**kwargs`.

    Args:
        dataset_name(str): Dataset name.
        description(str): dataset description. Defaults to None.
        is_version_enabled(bool): True if a dataset versioning is set,
            False otherwise. Defaults to False.
        is_public(bool): True if a dataset is source from a public bucket, False
            otherwise. Defaults to False.
        external_path(str): AWS S3 or Google Cloud Storage bucket URL. Defaults
            to None.
        aws_role_arn(str): AWS Role ARN to access S3. Defaults to None.
        version_path(str): Versioning bucket path. Defaults to None.

    Example:
        ```python
        vessl.create_dataset(
            dataset_name="mnist",
            is_public=True,
            external_path="s3://savvihub-public-apne2/mnist"
        )
        ```
    """
    if external_path is None:
        return _create_dataset_local(dataset_name, is_version_enabled, description, **kwargs)

    if external_path.startswith(DATASET_PATH_SCHEME_S3):
        return _create_dataset_s3(
            dataset_name,
            is_version_enabled,
            is_public,
            description,
            external_path,
            aws_role_arn,
            version_path,
            **kwargs,
        )

    if external_path.startswith(DATASET_PATH_SCHEME_GS):
        return _create_dataset_gs(
            dataset_name,
            is_version_enabled,
            is_public,
            description,
            external_path,
            version_path,
            **kwargs,
        )

    raise InvalidDatasetError("Invalid path scheme. Must be either s3:// or gs://.")


def list_dataset_volume_files(
    dataset_name: str,
    need_download_url: bool = False,
    path: str = "",
    recursive: bool = False,
    **kwargs,
) -> List[StorageFile]:
    """List dataset volume files in the default organization. If you want to
    override the default organization, then pass `organization_name` as
    `**kwargs`.

    Args:
        dataset_name(str): Dataset name.
        need_download_url(bool): True if you need a download URL, False
            otherwise. Defaults to False.
        path(str): Directory path to list. Defaults to root(""),
        recursive(bool): True if list files recursively, False otherwise.
            Defaults to False.

    Example:
        ```python
        vessl.list_dataset_volume_files(
            dataset_name="mnist",
            recursive=True,
        )
        ```
    """
    dataset = read_dataset(dataset_name, **kwargs)
    return list_volume_files(dataset.volume_id, need_download_url, path, recursive)


def upload_dataset_volume_file(
    dataset_name: str,
    source_path: str,
    dest_path: str,
    **kwargs,
) -> None:
    """Upload file to the dataset. If you want to override the default
    organization, then pass `organization_name` as `**kwargs`.

    Args:
        dataset_name(str): Dataset name.
        source_path(str): Local source path.
        dest_path(str): Destination path within the dataset.

    Example:
        ```python
        vessl.upload_dataset_volume_file(
            dataset_name="mnist",
            source_path="test.csv",
            dest_path="train",
        )
        ```
    """
    dataset = read_dataset(dataset_name, **kwargs)
    return copy_volume_file(
        source_volume_id=None,
        source_path=source_path,
        dest_volume_id=dataset.volume_id,
        dest_path=dest_path,
    )


def download_dataset_volume_file(
    dataset_name: str,
    source_path: str,
    dest_path: str,
    **kwargs,
) -> None:
    """Download file from the dataset. If you want to override the default
    organization, then pass `organization_name` as `**kwargs`.

    Args:
        dataset_name(str): Dataset name.
        source_path(str): Source path within the dataset.
        dest_path(str): Local destination path.

    Example:
        ```python
        vessl.download_dataset_volume_file(
            dataset_name="mnist",
            source_path="train/test.csv",
            dest_path=".",
        )
        ```
    """
    dataset = read_dataset(dataset_name, **kwargs)
    if dataset.source.provider == "s3":
        return copy_volume_file(
            source_volume_id=dataset.volume_id,
            source_path=source_path,
            dest_volume_id=None,
            dest_path=dest_path,
        )
    elif dataset.source.provider == "gs":
        return _copy_volume_file_remote_to_local(
            dataset.volume_id,
            source_path,
            dest_path,
        )
    else:
        print("Cannot download from on-premise dataset.")


def copy_dataset_volume_file(
    dataset_name: str,
    source_path: str,
    dest_path: str,
    **kwargs,
) -> None:
    """Copy files within the same dataset. Noted that this is not supported for
    externally sourced datasets like S3 or GCS. If you want to override the
    default organization, then pass `organization_name` as `**kwargs`.

    Args:
        dataset_name(str): Dataset name.
        source_path(str): Source path within the dataset.
        dest_path(str): Local destination path.

    Example:
        ```python
        vessl.download_dataset_volume_file(
            dataset_name="mnist",
            source_path="train/test.csv",
            dest_path="test/test.csv",
        )
        ```
    """
    dataset = read_dataset(dataset_name, **kwargs)
    return copy_volume_file(
        source_volume_id=dataset.volume_id,
        source_path=source_path,
        dest_volume_id=dataset.volume_id,
        dest_path=dest_path,
    )


def delete_dataset_volume_file(
    dataset_name: str,
    path: str,
    **kwargs,
) -> List[StorageFile]:
    """Delete the dataset volume file. Noted that this is not supported for
    externally sourced datasets like S3 or GCS. If you want to override the
    default organization, then pass `organization_name` as `**kwargs`.

    Args:
        dataset_name(str): Dataset name.
        path(str): File path.

    Example:
        ```python
        vessl.delete_dataset_volume_file(
            dataset_name="mnist",
            path="train/test.csv",
        )
        ```
    """
    dataset = read_dataset(dataset_name, **kwargs)
    return delete_volume_file(dataset.volume_id, path)



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from openapi_client.configuration import Configuration


class ResponseReducedPipelineStep(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'created_dt': 'datetime',
        'description': 'str',
        'external_service_spec': 'ResponsePipelineStepExternalServiceSpec',
        'id': 'int',
        'if_spec': 'ResponsePipelineStepIfSpec',
        'jupyter_visualization_spec': 'ResponseReducedPipelineStepJupyterVisualizationSpec',
        'manual_input_spec': 'ResponsePipelineStepManualInputSpec',
        'run_spec': 'ResponseReducedPipelineStepRunSpec',
        'step_key': 'str',
        'title': 'str',
        'type': 'str'
    }

    attribute_map = {
        'created_dt': 'created_dt',
        'description': 'description',
        'external_service_spec': 'external_service_spec',
        'id': 'id',
        'if_spec': 'if_spec',
        'jupyter_visualization_spec': 'jupyter_visualization_spec',
        'manual_input_spec': 'manual_input_spec',
        'run_spec': 'run_spec',
        'step_key': 'step_key',
        'title': 'title',
        'type': 'type'
    }

    def __init__(self, created_dt=None, description=None, external_service_spec=None, id=None, if_spec=None, jupyter_visualization_spec=None, manual_input_spec=None, run_spec=None, step_key=None, title=None, type=None, local_vars_configuration=None):  # noqa: E501
        """ResponseReducedPipelineStep - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._created_dt = None
        self._description = None
        self._external_service_spec = None
        self._id = None
        self._if_spec = None
        self._jupyter_visualization_spec = None
        self._manual_input_spec = None
        self._run_spec = None
        self._step_key = None
        self._title = None
        self._type = None
        self.discriminator = None

        self.created_dt = created_dt
        if description is not None:
            self.description = description
        if external_service_spec is not None:
            self.external_service_spec = external_service_spec
        self.id = id
        if if_spec is not None:
            self.if_spec = if_spec
        if jupyter_visualization_spec is not None:
            self.jupyter_visualization_spec = jupyter_visualization_spec
        if manual_input_spec is not None:
            self.manual_input_spec = manual_input_spec
        if run_spec is not None:
            self.run_spec = run_spec
        self.step_key = step_key
        self.title = title
        self.type = type

    @property
    def created_dt(self):
        """Gets the created_dt of this ResponseReducedPipelineStep.  # noqa: E501


        :return: The created_dt of this ResponseReducedPipelineStep.  # noqa: E501
        :rtype: datetime
        """
        return self._created_dt

    @created_dt.setter
    def created_dt(self, created_dt):
        """Sets the created_dt of this ResponseReducedPipelineStep.


        :param created_dt: The created_dt of this ResponseReducedPipelineStep.  # noqa: E501
        :type created_dt: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_dt is None:  # noqa: E501
            raise ValueError("Invalid value for `created_dt`, must not be `None`")  # noqa: E501

        self._created_dt = created_dt

    @property
    def description(self):
        """Gets the description of this ResponseReducedPipelineStep.  # noqa: E501


        :return: The description of this ResponseReducedPipelineStep.  # noqa: E501
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """Sets the description of this ResponseReducedPipelineStep.


        :param description: The description of this ResponseReducedPipelineStep.  # noqa: E501
        :type description: str
        """

        self._description = description

    @property
    def external_service_spec(self):
        """Gets the external_service_spec of this ResponseReducedPipelineStep.  # noqa: E501


        :return: The external_service_spec of this ResponseReducedPipelineStep.  # noqa: E501
        :rtype: ResponsePipelineStepExternalServiceSpec
        """
        return self._external_service_spec

    @external_service_spec.setter
    def external_service_spec(self, external_service_spec):
        """Sets the external_service_spec of this ResponseReducedPipelineStep.


        :param external_service_spec: The external_service_spec of this ResponseReducedPipelineStep.  # noqa: E501
        :type external_service_spec: ResponsePipelineStepExternalServiceSpec
        """

        self._external_service_spec = external_service_spec

    @property
    def id(self):
        """Gets the id of this ResponseReducedPipelineStep.  # noqa: E501


        :return: The id of this ResponseReducedPipelineStep.  # noqa: E501
        :rtype: int
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this ResponseReducedPipelineStep.


        :param id: The id of this ResponseReducedPipelineStep.  # noqa: E501
        :type id: int
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def if_spec(self):
        """Gets the if_spec of this ResponseReducedPipelineStep.  # noqa: E501


        :return: The if_spec of this ResponseReducedPipelineStep.  # noqa: E501
        :rtype: ResponsePipelineStepIfSpec
        """
        return self._if_spec

    @if_spec.setter
    def if_spec(self, if_spec):
        """Sets the if_spec of this ResponseReducedPipelineStep.


        :param if_spec: The if_spec of this ResponseReducedPipelineStep.  # noqa: E501
        :type if_spec: ResponsePipelineStepIfSpec
        """

        self._if_spec = if_spec

    @property
    def jupyter_visualization_spec(self):
        """Gets the jupyter_visualization_spec of this ResponseReducedPipelineStep.  # noqa: E501


        :return: The jupyter_visualization_spec of this ResponseReducedPipelineStep.  # noqa: E501
        :rtype: ResponseReducedPipelineStepJupyterVisualizationSpec
        """
        return self._jupyter_visualization_spec

    @jupyter_visualization_spec.setter
    def jupyter_visualization_spec(self, jupyter_visualization_spec):
        """Sets the jupyter_visualization_spec of this ResponseReducedPipelineStep.


        :param jupyter_visualization_spec: The jupyter_visualization_spec of this ResponseReducedPipelineStep.  # noqa: E501
        :type jupyter_visualization_spec: ResponseReducedPipelineStepJupyterVisualizationSpec
        """

        self._jupyter_visualization_spec = jupyter_visualization_spec

    @property
    def manual_input_spec(self):
        """Gets the manual_input_spec of this ResponseReducedPipelineStep.  # noqa: E501


        :return: The manual_input_spec of this ResponseReducedPipelineStep.  # noqa: E501
        :rtype: ResponsePipelineStepManualInputSpec
        """
        return self._manual_input_spec

    @manual_input_spec.setter
    def manual_input_spec(self, manual_input_spec):
        """Sets the manual_input_spec of this ResponseReducedPipelineStep.


        :param manual_input_spec: The manual_input_spec of this ResponseReducedPipelineStep.  # noqa: E501
        :type manual_input_spec: ResponsePipelineStepManualInputSpec
        """

        self._manual_input_spec = manual_input_spec

    @property
    def run_spec(self):
        """Gets the run_spec of this ResponseReducedPipelineStep.  # noqa: E501


        :return: The run_spec of this ResponseReducedPipelineStep.  # noqa: E501
        :rtype: ResponseReducedPipelineStepRunSpec
        """
        return self._run_spec

    @run_spec.setter
    def run_spec(self, run_spec):
        """Sets the run_spec of this ResponseReducedPipelineStep.


        :param run_spec: The run_spec of this ResponseReducedPipelineStep.  # noqa: E501
        :type run_spec: ResponseReducedPipelineStepRunSpec
        """

        self._run_spec = run_spec

    @property
    def step_key(self):
        """Gets the step_key of this ResponseReducedPipelineStep.  # noqa: E501


        :return: The step_key of this ResponseReducedPipelineStep.  # noqa: E501
        :rtype: str
        """
        return self._step_key

    @step_key.setter
    def step_key(self, step_key):
        """Sets the step_key of this ResponseReducedPipelineStep.


        :param step_key: The step_key of this ResponseReducedPipelineStep.  # noqa: E501
        :type step_key: str
        """
        if self.local_vars_configuration.client_side_validation and step_key is None:  # noqa: E501
            raise ValueError("Invalid value for `step_key`, must not be `None`")  # noqa: E501

        self._step_key = step_key

    @property
    def title(self):
        """Gets the title of this ResponseReducedPipelineStep.  # noqa: E501


        :return: The title of this ResponseReducedPipelineStep.  # noqa: E501
        :rtype: str
        """
        return self._title

    @title.setter
    def title(self, title):
        """Sets the title of this ResponseReducedPipelineStep.


        :param title: The title of this ResponseReducedPipelineStep.  # noqa: E501
        :type title: str
        """
        if self.local_vars_configuration.client_side_validation and title is None:  # noqa: E501
            raise ValueError("Invalid value for `title`, must not be `None`")  # noqa: E501

        self._title = title

    @property
    def type(self):
        """Gets the type of this ResponseReducedPipelineStep.  # noqa: E501


        :return: The type of this ResponseReducedPipelineStep.  # noqa: E501
        :rtype: str
        """
        return self._type

    @type.setter
    def type(self, type):
        """Sets the type of this ResponseReducedPipelineStep.


        :param type: The type of this ResponseReducedPipelineStep.  # noqa: E501
        :type type: str
        """
        if self.local_vars_configuration.client_side_validation and type is None:  # noqa: E501
            raise ValueError("Invalid value for `type`, must not be `None`")  # noqa: E501

        self._type = type

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ResponseReducedPipelineStep):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ResponseReducedPipelineStep):
            return True

        return self.to_dict() != other.to_dict()



from typing import List

from openapi_client.models import OrmKernelImage, ResponseKernelImage
from vessl import vessl_api
from vessl.organization import _get_organization_name


def read_kernel_image(image_id: int) -> ResponseKernelImage:
    """Read the kernel image.

    Args:
        image_id(int): Image ID.

    Example:
        ```python
        vessl.read_kernel_image(
            image_id=1,
        )
        ```
    """
    return vessl_api.kernel_image_read_api(image_id=image_id)


def list_kernel_images(**kwargs) -> List[OrmKernelImage]:
    """List kernel images in the default organization. If you
    want to override the default organization, then pass `organization_name` as
    `**kwargs`.

    Example:
        ```python
        vessl.list_kernel_images()
        ```
    """
    return vessl_api.kernel_image_list_api(
        organization_name=_get_organization_name(**kwargs),
    ).results



from datetime import datetime, timezone
from importlib import import_module

import shortuuid
import timeago

from vessl.util import logger
from vessl.util.exception import ImportPackageError


def parse_time_to_ago(dt: datetime):
    if not dt:
        return "N/A"
    return timeago.format(dt, datetime.now(timezone.utc))


def safe_cast(val, to_type, default=None):
    try:
        return to_type(val)
    except (ValueError, TypeError):
        return default


def get_module(name, required=None):
    try:
        return import_module(name)
    except ImportError:
        msg = f"Error importing optional module {name}"
        if required:
            logger.warn(msg)
            raise ImportPackageError(f"{required}")


def generate_uuid():
    generated_uuid = shortuuid.ShortUUID(alphabet=list("0123456789abcdefghijklmnopqrstuvwxyz"))
    return generated_uuid.random(8)


def is_processor_m1():
    from cpuinfo import get_cpu_info

    return "m1" in get_cpu_info().get("brand_raw").lower()


def remove_prefix(text, prefix):
    if text.startswith(prefix):
        return text[len(prefix) :]
    return text



import base64
import socket
import subprocess
from platform import uname
from shutil import which
from time import sleep
from typing import List, Optional, Tuple

import docker
from kubernetes import client as k8sclient
from kubernetes import config
from kubernetes.config import KUBE_CONFIG_DEFAULT_LOCATION

from openapi_client.models import (
    ClusterUpdateAPIInput,
    ResponseKernelClusterInfo,
    ResponseKernelClusterNodeInfo,
)
from vessl import vessl_api
from vessl.organization import _get_organization_name
from vessl.util import logger
from vessl.util.config import DEFAULT_VESSL_DIR
from vessl.util.constant import CLUSTER_MODE_MULTI_NODE, CLUSTER_MODE_SINGLE
from vessl.util.exception import (
    ClusterNotFoundError,
    InvalidKernelClusterError,
    VesslApiException,
    VesslRuntimeException,
)


def create_cluster(
    cluster_name: str,
    kubernetes_namespace: str = "vessl",
    extra_helm_values: List[str] = [],
    kubeconfig_path: str = "",
    provider: str = "on-premise",
    **kwargs,
) -> ResponseKernelClusterInfo:
    """Create a VESSL cluster by installing VESSL agent to given Kubernetes
    namespace. If you want to override the default organization, then pass
    `organization_name` as `**kwargs`.

    Args:
        cluster_name(str): Cluster name.
        kubernetes_namespace(str): Kubernetes namespace to install VESSL agent.
            defaults to "vessl".
        extra_helm_values(list[str]): Helm values to pass to cluster install
            command. See https://github.com/vessl-ai/cluster-resources/blob/main/helm-chart/values.yaml
            for available Helm values.
        kubeconfig_path(str): Path to kubeconfig file.
        provider(str): Cluster provider. defaults to "on-premise"

    Example:
        ```python
        vessl.install_cluster(
            cluster_name="seoul-cluster",
        )
        ```
    """
    agent_access_token = vessl_api.custom_cluster_key_api(
        organization_name=_get_organization_name(**kwargs),
    ).access_token

    logger.debug(f"cluster name: '{cluster_name}'")
    logger.debug(f"access token: '{agent_access_token}'")

    subprocess.check_call(
        [
            "helm",
            "repo",
            "add",
            "vessl",
            "https://vessl-ai.github.io/cluster-resources/helm-chart",
        ]
    )
    subprocess.check_call(["helm", "repo", "update"])

    cluster_install_command = [
        "helm",
        "install",
        "vessl",
        "vessl/cluster-resources",
        "--kubeconfig",
        kubeconfig_path,
        "--namespace",
        kubernetes_namespace,
        "--create-namespace",
        "--set",
        f"agent.clusterName={cluster_name}",
        "--set",
        f"agent.accessToken={agent_access_token}",
        "--set",
        f"agent.providerType={provider}",
    ]

    if provider == "gcp":
        cluster_install_command.extend(
            [
                "--set",
                f"agent.localStorageClassName=standard-rwo",
            ]
        )

    if provider != "on-premise":  # Disable local path provisioner if provider is cloud service
        cluster_install_command.extend(
            [
                "--set",
                f"localPathProvisioner.enabled=false",
            ]
        )

    for helm_value in extra_helm_values:
        cluster_install_command.extend(["--set", helm_value])

    # dcgm patch
    v1 = k8sclient.CoreV1Api()
    # @XXX(seokju) rough assumption: k0s user will install k0s nodes
    # if this becomes a problem, patch chart and cli.
    node = v1.list_node().items[0]
    if "k0s" in node.status.node_info.kubelet_version:
        cluster_install_command.extend(
            [
                "--set",
                ".metricsExporters.dcgmExporter.podResourcePath=/var/lib/k0s/kubelet/pod-resources",
            ]
        )

    subprocess.check_call(cluster_install_command)

    logger.warn(
        "VESSL cluster agent installed. Waiting for the agent to be connected with VESSL..."
    )
    for attempts in range(18):
        sleep(10)
        try:
            return vessl_api.custom_cluster_connectivity_api(
                organization_name=_get_organization_name(**kwargs),
                agent_access_token=agent_access_token,
            )
        except VesslApiException as e:
            if e.code == "NotFound":
                continue
            raise e

    raise ClusterNotFoundError(
        "Timeout for checking agent connection; Please check agent log for more information."
    )


def _check_cluster_install_dependencies(mode: str):
    """Check if dependencies required for cluster creation has installed on
    the machine."""
    if mode == CLUSTER_MODE_SINGLE:
        try:
            logger.debug("Cluster install: checking if Docker daemon is running")
            docker.from_env()
        except docker.errors.DockerException as e:
            raise VesslRuntimeException(
                f"Error response from Docker daemon: {repr(e)}; Please install Docker to run VESSL on your machine."
            )

    logger.debug("Cluster install: checking if Helm binary exists in local")
    if which("helm") is None:
        raise VesslRuntimeException(
            "Command 'helm' not found; Please install Helm to install VESSL cluster agent: https://helm.sh/"
        )


def _load_cluster_kubeconfig(
    mode: str,
) -> Tuple[bool, Tuple[Optional[str], Optional[str]]]:
    """Find and load kubeconfig for VESSL cluster installation.

    Returns:
        True, if valid kubernetes config has loaded
        False, if no valid kubernetes config has found
    """
    cluster_kubeconfig = _get_cluster_kubeconfig()
    config_files = [cluster_kubeconfig] if cluster_kubeconfig else []
    if mode == CLUSTER_MODE_MULTI_NODE:
        config_files.append("/var/lib/k0s/pki/admin.conf")
    config_files.append(KUBE_CONFIG_DEFAULT_LOCATION)

    for config_file in config_files:
        try:
            config.load_kube_config(config_file)
            _, current_context = config.list_kube_config_contexts(config_file)
            logger.info(
                f"Using kubeconfig from file: {config_file} / context: {current_context['name']}"
            )
            return True, (config_file, current_context["name"])
        except config.config_exception.ConfigException as e:
            logger.warn(f"Cluster install: load kubeconfig from {config_file} failed: {repr(e)}")

    return False, (None, None)


def _install_k0s_in_docker() -> Tuple[str, str]:
    """Install k0s-in-Docker on local machine."""
    logger.info(f"Installing k0s-in-docker ...")
    k0s_config = """apiVersion: k0s.k0sproject.io/v1beta1
kind: ClusterConfig
metadata:
  name: k0s
spec:
  api:
    extraArgs:
      service-node-port-range: 32639-32767
"""
    kwargs = {
        "name": "vessl-k0s",
        "environment": {"K0S_CONFIG": k0s_config},
        "hostname": f"vessl-k0s-{socket.gethostname()}",
        "ports": {
            "6443/tcp": 6443,
        },
        "privileged": True,
        "volumes": {
            "/var/lib/k0s": {
                "bind": "/var/lib/k0s",
                "mode": "rw",
            },
        },
    }
    for i in range(32639, 32768):
        kwargs["ports"][f"{i}/tcp"] = i

    # Add extra arguments for arm64 machines
    if uname().machine == "arm64":
        logger.debug("System is using CPU with arm64 architecture; adding additional kwargs")
        kwargs["environment"]["ETCD_UNSUPPORTED_ARCH"] = "arm64"

    client = docker.from_env()

    # Add extra arguments for docker container running on Docker Desktop
    if client.info().get("OperatingSystem", "") == "Docker Desktop":
        logger.debug("Docker is running on Docker Desktop; adding additional kwargs")
        kwargs["cgroupns"] = "host"
        kwargs["volumes"]["/sys/fs/cgroup"] = {
            "bind": "/sys/fs/cgroup",
            "mode": "rw",
        }

    # Run k0s container
    container = client.containers.run(
        detach=True,
        image="docker.io/k0sproject/k0s:v1.23.9-k0s.0",
        command="k0s controller --single --no-taints --config /etc/k0s/config.yaml",
        **kwargs,
    )

    # Wait for the container to be ready
    node_ready = False
    trial_limit = 10
    for trial in range(trial_limit):
        logger.info(f"Checking for the k0s-in-docker node to be ready... ({trial+1}/{trial_limit})")
        res = container.exec_run(
            "bash -c 'k0s kubectl get nodes --no-headers=true | grep -v NotReady | wc -l'",
            stream=False,
            demux=True,
        )
        output, err = res.output
        if output:
            logger.debug(output.decode("utf-8"))
            nodes = output.decode("utf-8").strip()
            if nodes == "1":
                node_ready = True
                break
        if err:
            logger.debug(err.decode("utf-8"))
        sleep(10)
    if not node_ready:
        container.kill()
        raise VesslRuntimeException("Timeout while waiting for k0s-in-docker node. Please retry.")

    # Get kubeconfig, keep at /opt/vessl/cluster_kubeconfig.yaml for later
    logger.debug("Acquiring kubeconfig")
    res = container.exec_run("cat /var/lib/k0s/pki/admin.conf", stream=False, demux=True)
    output, _ = res.output
    kubeconfig = output.decode("utf-8")
    kubeconfig_filename = f"{DEFAULT_VESSL_DIR}/cluster_kubeconfig.yaml"
    with open(kubeconfig_filename, "w") as f:
        f.write(kubeconfig)

    vessl_api.config_loader.cluster_kubeconfig = kubeconfig_filename
    config.load_kube_config(kubeconfig_filename)
    _, current_context = config.list_kube_config_contexts(kubeconfig_filename)
    logger.info(f"k0s-in-docker is ready! Saved kubeconfig at {kubeconfig_filename} for later use.")
    return kubeconfig_filename, current_context["name"]


def _check_existing_vessl_installation(**kwargs) -> ResponseKernelClusterInfo:
    """Check if there is existing vessl installation in current Kubernetes
    cluster."""

    # Find secret with name=vessl-agent, annotation=vessl
    try:
        v1 = k8sclient.CoreV1Api()
        vessl_secret: k8sclient.V1Secret = v1.read_namespaced_secret(
            name="vessl-agent", namespace="vessl"
        )
    except k8sclient.ApiException as e:
        if e.status != 404:
            raise VesslRuntimeException("Error while communicating with kubernetes.")
        return None

    # VESSL agent token is found - there is existing cluster in current kubernetes cluster
    logger.warn("Existing VESSL cluster installation found! getting cluster information...")
    vessl_access_token = vessl_secret.data.get("access-token")
    if vessl_access_token is None:
        raise InvalidKernelClusterError(
            "Kubernetes secret for VESSL found, but secret does not have access-token"
        )
    vessl_access_token = base64.b64decode(vessl_access_token)
    logger.debug(f"access token: {vessl_access_token}")
    return vessl_api.custom_cluster_connectivity_api(
        organization_name=_get_organization_name(**kwargs),
        agent_access_token=vessl_access_token,
    )


def read_cluster(cluster_name: str, **kwargs) -> ResponseKernelClusterInfo:
    """Read cluster in the default organization. If you want to override the
    default organization, then pass `organization_name` as `**kwargs`.

    Args:
        cluster_name(str): Cluster name.

    Example:
        ```python
        vessl.read_cluster(
            cluster_name="seoul-cluster",
        )
        ```
    """
    kernel_clusters = list_clusters(**kwargs)
    kernel_clusters = {x.name: x for x in kernel_clusters}

    if cluster_name not in kernel_clusters:
        raise InvalidKernelClusterError(f"Kernel cluster not found: {cluster_name}")
    return kernel_clusters[cluster_name]


def list_clusters(**kwargs) -> List[ResponseKernelClusterInfo]:
    """List clusters in the default organization. If you want to override the
    default organization, then pass `organization_name` as `**kwargs`.

    Example:
        ```python
        vessl.list_clusters()
        ```
    """
    return vessl_api.cluster_list_api(
        organization_name=_get_organization_name(**kwargs),
    ).clusters


def delete_cluster(cluster_id: int, **kwargs) -> object:
    """Delete custom cluster in the default organization. If you want to
    override the default organization, then pass `organization_name` as
    `**kwargs`.

    Args:
        cluster_id(int): Cluster ID.

    Example:
        ```python
        vessl.delete_cluster(
            cluster_id=1,
        )
        ```
    """
    return vessl_api.cluster_delete_api(
        cluster_id=cluster_id,
        organization_name=_get_organization_name(**kwargs),
    )


def rename_cluster(cluster_id: int, new_cluster_name: str, **kwargs) -> ResponseKernelClusterInfo:
    """Rename custom cluster in the default organization. If you want to
    override the default organization, then pass `organization_name` as
    `**kwargs`.

    Args:
        cluster_id(int): Cluster ID.
        new_cluster_name(str): Cluster name to change.

    Example:
        ```python
        vessl.rename_cluster(
            cluster_id=1,
            new_cluster_name="seoul-cluster-2",
        )
        ```
    """
    return vessl_api.cluster_update_api(
        cluster_id=cluster_id,
        organization_name=_get_organization_name(**kwargs),
        custom_cluster_update_api_input=ClusterUpdateAPIInput(
            name=new_cluster_name,
        ),
    )


def list_cluster_nodes(cluster_id: int, **kwargs) -> List[ResponseKernelClusterNodeInfo]:
    """List custom cluster nodes in the default organization. If you want to
    override the default organization, then pass `organization_name` as
    `**kwargs`.

    Args:
        cluster_id(int): Cluster ID.

    Example:
        ```python
        vessl.list_cluster_nodes(
            cluster_id=1,
        )
        ```
    """
    return vessl_api.custom_cluster_node_list_api(
        cluster_id=cluster_id,
        organization_name=_get_organization_name(**kwargs),
    ).nodes


def _get_cluster_kubeconfig(**kwargs) -> str:
    cluster_kubeconfig = kwargs.get("cluster_kubeconfig")
    if cluster_kubeconfig is not None:
        return cluster_kubeconfig
    if vessl_api.cluster_kubeconfig is not None:
        return vessl_api.cluster_kubeconfig

    return ""



import os
import tempfile
from test.conftest import USE_MOCK

import pytest

import vessl
from vessl.util.file_object import UploadableS3Object
from vessl.util.random import random_string


@pytest.mark.skipif(USE_MOCK, reason="Does not run if mocking is used.")
class TestVolume:
    file_name = random_string()

    @pytest.mark.order(index=1)
    def test_create_volume_file(self):
        volume_id = vessl.vessl_api.project.volume_id
        skeleton = vessl.create_volume_file(volume_id, False, self.file_name)

        # create an actual file on local
        local_path = os.path.join(tempfile.gettempdir(), self.file_name)
        with open(local_path, "w") as f:
            f.write("dummy")

        UploadableS3Object(
            local_path,
            skeleton.upload_url.federation_token.bucket,
            skeleton.upload_url.federation_token.key,
            skeleton.upload_url.federation_token.token,
        ).upload()

        # remove the local file
        os.remove(local_path)

    def test_read_volume_file(self):
        volume_id = vessl.vessl_api.project.volume_id
        vessl.read_volume_file(volume_id, self.file_name)

    def test_list_volume_files(self):
        volume_id = vessl.vessl_api.project.volume_id
        vessl.list_volume_files(volume_id)

    @pytest.mark.order(index=-1)
    def test_delete_volume_file(self):
        volume_id = vessl.vessl_api.project.volume_id
        vessl.delete_volume_file(volume_id, path=self.file_name)



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from openapi_client.configuration import Configuration


class ResponseBillingHistoryInfo(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'amount': 'float',
        'created_by': 'ResponseSimpleUser',
        'created_dt': 'datetime',
        'discounted_amount': 'float',
        'experiment': 'ResponseSimpleExperiment',
        'model_service_revision': 'ResponseModelServiceRevisionInfo',
        'pipeline_step_execution': 'ResponseReducedPipelineStepExecution',
        'project': 'ResponseSimpleProject',
        'service': 'ResponseSimpleServiceInfo',
        'status': 'str',
        'type': 'str',
        'workspace': 'ResponseSimpleWorkspace'
    }

    attribute_map = {
        'amount': 'amount',
        'created_by': 'created_by',
        'created_dt': 'created_dt',
        'discounted_amount': 'discounted_amount',
        'experiment': 'experiment',
        'model_service_revision': 'model_service_revision',
        'pipeline_step_execution': 'pipeline_step_execution',
        'project': 'project',
        'service': 'service',
        'status': 'status',
        'type': 'type',
        'workspace': 'workspace'
    }

    def __init__(self, amount=None, created_by=None, created_dt=None, discounted_amount=None, experiment=None, model_service_revision=None, pipeline_step_execution=None, project=None, service=None, status=None, type=None, workspace=None, local_vars_configuration=None):  # noqa: E501
        """ResponseBillingHistoryInfo - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._amount = None
        self._created_by = None
        self._created_dt = None
        self._discounted_amount = None
        self._experiment = None
        self._model_service_revision = None
        self._pipeline_step_execution = None
        self._project = None
        self._service = None
        self._status = None
        self._type = None
        self._workspace = None
        self.discriminator = None

        self.amount = amount
        self.created_by = created_by
        self.created_dt = created_dt
        self.discounted_amount = discounted_amount
        if experiment is not None:
            self.experiment = experiment
        if model_service_revision is not None:
            self.model_service_revision = model_service_revision
        if pipeline_step_execution is not None:
            self.pipeline_step_execution = pipeline_step_execution
        self.project = project
        if service is not None:
            self.service = service
        self.status = status
        self.type = type
        if workspace is not None:
            self.workspace = workspace

    @property
    def amount(self):
        """Gets the amount of this ResponseBillingHistoryInfo.  # noqa: E501


        :return: The amount of this ResponseBillingHistoryInfo.  # noqa: E501
        :rtype: float
        """
        return self._amount

    @amount.setter
    def amount(self, amount):
        """Sets the amount of this ResponseBillingHistoryInfo.


        :param amount: The amount of this ResponseBillingHistoryInfo.  # noqa: E501
        :type amount: float
        """
        if self.local_vars_configuration.client_side_validation and amount is None:  # noqa: E501
            raise ValueError("Invalid value for `amount`, must not be `None`")  # noqa: E501

        self._amount = amount

    @property
    def created_by(self):
        """Gets the created_by of this ResponseBillingHistoryInfo.  # noqa: E501


        :return: The created_by of this ResponseBillingHistoryInfo.  # noqa: E501
        :rtype: ResponseSimpleUser
        """
        return self._created_by

    @created_by.setter
    def created_by(self, created_by):
        """Sets the created_by of this ResponseBillingHistoryInfo.


        :param created_by: The created_by of this ResponseBillingHistoryInfo.  # noqa: E501
        :type created_by: ResponseSimpleUser
        """
        if self.local_vars_configuration.client_side_validation and created_by is None:  # noqa: E501
            raise ValueError("Invalid value for `created_by`, must not be `None`")  # noqa: E501

        self._created_by = created_by

    @property
    def created_dt(self):
        """Gets the created_dt of this ResponseBillingHistoryInfo.  # noqa: E501


        :return: The created_dt of this ResponseBillingHistoryInfo.  # noqa: E501
        :rtype: datetime
        """
        return self._created_dt

    @created_dt.setter
    def created_dt(self, created_dt):
        """Sets the created_dt of this ResponseBillingHistoryInfo.


        :param created_dt: The created_dt of this ResponseBillingHistoryInfo.  # noqa: E501
        :type created_dt: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_dt is None:  # noqa: E501
            raise ValueError("Invalid value for `created_dt`, must not be `None`")  # noqa: E501

        self._created_dt = created_dt

    @property
    def discounted_amount(self):
        """Gets the discounted_amount of this ResponseBillingHistoryInfo.  # noqa: E501


        :return: The discounted_amount of this ResponseBillingHistoryInfo.  # noqa: E501
        :rtype: float
        """
        return self._discounted_amount

    @discounted_amount.setter
    def discounted_amount(self, discounted_amount):
        """Sets the discounted_amount of this ResponseBillingHistoryInfo.


        :param discounted_amount: The discounted_amount of this ResponseBillingHistoryInfo.  # noqa: E501
        :type discounted_amount: float
        """
        if self.local_vars_configuration.client_side_validation and discounted_amount is None:  # noqa: E501
            raise ValueError("Invalid value for `discounted_amount`, must not be `None`")  # noqa: E501

        self._discounted_amount = discounted_amount

    @property
    def experiment(self):
        """Gets the experiment of this ResponseBillingHistoryInfo.  # noqa: E501


        :return: The experiment of this ResponseBillingHistoryInfo.  # noqa: E501
        :rtype: ResponseSimpleExperiment
        """
        return self._experiment

    @experiment.setter
    def experiment(self, experiment):
        """Sets the experiment of this ResponseBillingHistoryInfo.


        :param experiment: The experiment of this ResponseBillingHistoryInfo.  # noqa: E501
        :type experiment: ResponseSimpleExperiment
        """

        self._experiment = experiment

    @property
    def model_service_revision(self):
        """Gets the model_service_revision of this ResponseBillingHistoryInfo.  # noqa: E501


        :return: The model_service_revision of this ResponseBillingHistoryInfo.  # noqa: E501
        :rtype: ResponseModelServiceRevisionInfo
        """
        return self._model_service_revision

    @model_service_revision.setter
    def model_service_revision(self, model_service_revision):
        """Sets the model_service_revision of this ResponseBillingHistoryInfo.


        :param model_service_revision: The model_service_revision of this ResponseBillingHistoryInfo.  # noqa: E501
        :type model_service_revision: ResponseModelServiceRevisionInfo
        """

        self._model_service_revision = model_service_revision

    @property
    def pipeline_step_execution(self):
        """Gets the pipeline_step_execution of this ResponseBillingHistoryInfo.  # noqa: E501


        :return: The pipeline_step_execution of this ResponseBillingHistoryInfo.  # noqa: E501
        :rtype: ResponseReducedPipelineStepExecution
        """
        return self._pipeline_step_execution

    @pipeline_step_execution.setter
    def pipeline_step_execution(self, pipeline_step_execution):
        """Sets the pipeline_step_execution of this ResponseBillingHistoryInfo.


        :param pipeline_step_execution: The pipeline_step_execution of this ResponseBillingHistoryInfo.  # noqa: E501
        :type pipeline_step_execution: ResponseReducedPipelineStepExecution
        """

        self._pipeline_step_execution = pipeline_step_execution

    @property
    def project(self):
        """Gets the project of this ResponseBillingHistoryInfo.  # noqa: E501


        :return: The project of this ResponseBillingHistoryInfo.  # noqa: E501
        :rtype: ResponseSimpleProject
        """
        return self._project

    @project.setter
    def project(self, project):
        """Sets the project of this ResponseBillingHistoryInfo.


        :param project: The project of this ResponseBillingHistoryInfo.  # noqa: E501
        :type project: ResponseSimpleProject
        """
        if self.local_vars_configuration.client_side_validation and project is None:  # noqa: E501
            raise ValueError("Invalid value for `project`, must not be `None`")  # noqa: E501

        self._project = project

    @property
    def service(self):
        """Gets the service of this ResponseBillingHistoryInfo.  # noqa: E501


        :return: The service of this ResponseBillingHistoryInfo.  # noqa: E501
        :rtype: ResponseSimpleServiceInfo
        """
        return self._service

    @service.setter
    def service(self, service):
        """Sets the service of this ResponseBillingHistoryInfo.


        :param service: The service of this ResponseBillingHistoryInfo.  # noqa: E501
        :type service: ResponseSimpleServiceInfo
        """

        self._service = service

    @property
    def status(self):
        """Gets the status of this ResponseBillingHistoryInfo.  # noqa: E501


        :return: The status of this ResponseBillingHistoryInfo.  # noqa: E501
        :rtype: str
        """
        return self._status

    @status.setter
    def status(self, status):
        """Sets the status of this ResponseBillingHistoryInfo.


        :param status: The status of this ResponseBillingHistoryInfo.  # noqa: E501
        :type status: str
        """
        if self.local_vars_configuration.client_side_validation and status is None:  # noqa: E501
            raise ValueError("Invalid value for `status`, must not be `None`")  # noqa: E501

        self._status = status

    @property
    def type(self):
        """Gets the type of this ResponseBillingHistoryInfo.  # noqa: E501


        :return: The type of this ResponseBillingHistoryInfo.  # noqa: E501
        :rtype: str
        """
        return self._type

    @type.setter
    def type(self, type):
        """Sets the type of this ResponseBillingHistoryInfo.


        :param type: The type of this ResponseBillingHistoryInfo.  # noqa: E501
        :type type: str
        """
        if self.local_vars_configuration.client_side_validation and type is None:  # noqa: E501
            raise ValueError("Invalid value for `type`, must not be `None`")  # noqa: E501

        self._type = type

    @property
    def workspace(self):
        """Gets the workspace of this ResponseBillingHistoryInfo.  # noqa: E501


        :return: The workspace of this ResponseBillingHistoryInfo.  # noqa: E501
        :rtype: ResponseSimpleWorkspace
        """
        return self._workspace

    @workspace.setter
    def workspace(self, workspace):
        """Sets the workspace of this ResponseBillingHistoryInfo.


        :param workspace: The workspace of this ResponseBillingHistoryInfo.  # noqa: E501
        :type workspace: ResponseSimpleWorkspace
        """

        self._workspace = workspace

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ResponseBillingHistoryInfo):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ResponseBillingHistoryInfo):
            return True

        return self.to_dict() != other.to_dict()



import functools
import inspect
import os
import time
import webbrowser

from openapi_client import ApiClient, APIV1Api, Configuration
from openapi_client.exceptions import ApiException
from openapi_client.models import (
    ResponseMyUser,
    ResponseOrganization,
    ResponseProjectInfo,
)
from vessl._version import __VERSION__
from vessl.util.config import VesslConfigLoader
from vessl.util.constant import (
    ACCESS_TOKEN_ENV_VAR,
    API_HOST,
    CLUSTER_KUBECONFIG_ENV_VAR,
    CREDENTIALS_FILE_ENV_VAR,
    DEFAULT_ORGANIZATION_ENV_VAR,
    DEFAULT_PROJECT_ENV_VAR,
    LOGIN_TIMEOUT_SECONDS,
    VESSL_LOG_LEVEL,
    VESSL_LOG_LEVEL_DEBUG,
    WEB_HOST,
)
from vessl.util.exception import (
    GitError,
    InvalidOrganizationError,
    InvalidProjectError,
    InvalidTokenError,
    VesslApiException,
    VesslException,
    VesslRuntimeException,
)
from vessl.util.git_local_repo import GitRepository


def raise_vessl_exception(f):
    @functools.wraps(f)
    def func(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except ApiException as e:
            raise VesslApiException.convert_api_exception(e) from None

    return func


class VesslApi(APIV1Api):
    """VESSL Client API"""

    def __init__(self, *, api_host=API_HOST, **kwargs):
        _api_configuration = Configuration(host=api_host)
        # Disable verify ssl (https://github.com/urllib3/urllib3/issues/1682#issuecomment-533311857) # noqa E501
        _api_configuration.verify_ssl = False
        _api_configuration.client_side_validation = False
        if VESSL_LOG_LEVEL == VESSL_LOG_LEVEL_DEBUG:
            _api_configuration.debug = True

        Configuration.set_default(_api_configuration)

        _api_client = ApiClient(configuration=_api_configuration)
        import urllib3

        urllib3.disable_warnings()

        for k, v in kwargs:
            _api_client.set_default_header(k, v)
        _api_client.set_default_header("X-Source", "sdk")
        _api_client.set_default_header("X-Version", __VERSION__)
        super().__init__(_api_client)

        # `predicate=inspect.isfunction` does not work for some reason.
        # For more info: https://stackoverflow.com/q/17019949
        for name, value in inspect.getmembers(self, predicate=inspect.ismethod):
            if name.endswith("api"):
                setattr(self, name, raise_vessl_exception(value))

        self.user: ResponseMyUser = None
        self.organization: ResponseOrganization = None
        self.project: ResponseProjectInfo = None
        self.default_git_repo: GitRepository = None

        self.config_loader = VesslConfigLoader()
        self.cluster_kubeconfig = self.config_loader.cluster_kubeconfig
        if self.config_loader.access_token is not None:
            self.api_client.set_default_header(
                "Authorization",
                f"Token {self.config_loader.access_token}",
            )

        try:
            if self.config_loader.default_organization is not None:
                self.set_organization(self.config_loader.default_organization)

            if self.config_loader.default_project is not None:
                self.set_project(self.config_loader.default_project)
        except VesslException:
            pass

        self._initialize_git_repo()

    def configure(
        self,
        *,
        access_token: str = None,
        organization_name: str = None,
        project_name: str = None,
        credentials_file: str = None,
        force_update_access_token: bool = False,
    ) -> None:
        """Configure VESSL Client API.

        Args:
            access_token(str): Access token to override. Defaults to
                `access_token` from `~/.vessl/config`.
            organization_name(str): Organization name to override. Defaults to
                `organization_name` from `~/.vessl/config`.
            project_name(str): Project name to override. Defaults to
                `project name` from `~/.vessl/config`.
            credentials_file(str): Defaults to None.
            force_update_access_token(bool): True if force update access token,
                False otherwise. Defaults to False.

        Example:
            ```python
            vessl.configure()
            ```
        """
        self.configure_access_token(access_token, credentials_file, force_update_access_token)
        self.configure_default_organization(organization_name, credentials_file)

        try:
            self.configure_default_project(project_name, credentials_file)
        except InvalidProjectError:
            # Project is not mandatory
            pass

        print(
            f"""Username: {self.user.username}
Email: {self.user.email}
Default organization: {self.organization.name}
Default project: {'N/A' if self.project is None else self.project.name}"""
        )

    def _initialize_git_repo(self) -> None:
        try:
            self.default_git_repo = GitRepository()
        except GitError:
            return

    def _get_new_access_token(self) -> str:
        cli_token = self.sign_in_cli_token_api().cli_token
        url = f"{WEB_HOST}/cli/grant-access?token={cli_token}"

        try:
            webbrowser.open(url)
        finally:
            print(
                f"Please grant CLI access from the URL below.\n" f"{url}\n\nWaiting...\n",
            )

        start_time = time.time()
        while time.time() - start_time < LOGIN_TIMEOUT_SECONDS:
            response = self.sign_in_cli_check_api(cli_token)
            if response.signin_success:
                return response.access_token
            time.sleep(3)

        raise TimeoutError("Login timeout. Please try again.")

    def find_access_token(
        self,
        access_token: str = None,
        credentials_file: str = None,
        force_update: bool = False,
        no_prompt: bool = False,
    ) -> str:
        if force_update:
            return self._get_new_access_token()

        if access_token is not None:
            return access_token

        if credentials_file is not None:
            return VesslConfigLoader(credentials_file).access_token

        if os.environ.get(ACCESS_TOKEN_ENV_VAR):
            return os.environ.get(ACCESS_TOKEN_ENV_VAR)

        if os.environ.get(CREDENTIALS_FILE_ENV_VAR):
            file = os.environ.get(CREDENTIALS_FILE_ENV_VAR)
            return VesslConfigLoader(file).access_token

        if self.config_loader.access_token:
            return self.config_loader.access_token

        if not no_prompt:
            return self._get_new_access_token()

        raise InvalidTokenError("No access token found.")

    def find_organization_name(
        self, organization_name: str = None, credentials_file: str = None
    ) -> str:
        if organization_name is not None:
            return organization_name

        if credentials_file is not None:
            return VesslConfigLoader(credentials_file).default_organization

        if os.environ.get(DEFAULT_ORGANIZATION_ENV_VAR):
            return os.environ.get(DEFAULT_ORGANIZATION_ENV_VAR)

        if os.environ.get(CREDENTIALS_FILE_ENV_VAR):
            file = os.environ.get(CREDENTIALS_FILE_ENV_VAR)
            return VesslConfigLoader(file).default_organization

        if self.config_loader.default_organization:
            return self.config_loader.default_organization

        raise InvalidOrganizationError("Please specify an organization.")

    def find_project_name(self, project_name: str = None, credentials_file: str = None) -> str:
        if project_name is not None:
            return project_name

        if credentials_file is not None:
            return VesslConfigLoader(credentials_file).default_project

        if os.environ.get(CREDENTIALS_FILE_ENV_VAR):
            file = os.environ.get(CREDENTIALS_FILE_ENV_VAR)
            return VesslConfigLoader(file).default_project

        if os.environ.get(DEFAULT_PROJECT_ENV_VAR):
            return os.environ.get(DEFAULT_PROJECT_ENV_VAR)

        if self.config_loader.default_project:
            return self.config_loader.default_project

        raise InvalidProjectError("Please specify a project.")

    def find_cluster_kubeconfig(
        self, kubeconfig_path: str = None, credentials_file: str = None
    ) -> str:
        if kubeconfig_path is not None:
            return kubeconfig_path

        if credentials_file is not None:
            return VesslConfigLoader(credentials_file).cluster_kubeconfig

        if os.environ.get(CLUSTER_KUBECONFIG_ENV_VAR):
            return os.environ.get(CLUSTER_KUBECONFIG_ENV_VAR)

        if os.environ.get(CREDENTIALS_FILE_ENV_VAR):
            file = os.environ.get(CREDENTIALS_FILE_ENV_VAR)
            return VesslConfigLoader(file).cluster_kubeconfig

        if self.config_loader.cluster_kubeconfig:
            return self.config_loader.cluster_kubeconfig

        raise VesslRuntimeException("No Kubernetes cluster found.")

    def set_access_token(
        self,
        access_token: str = None,
        credentials_file: str = None,
        force_update: bool = False,
        no_prompt: bool = False,
    ) -> str:
        access_token = self.find_access_token(
            access_token,
            credentials_file,
            force_update,
            no_prompt,
        )
        self.api_client.set_default_header(
            "Authorization",
            f"Token {access_token}",
        )

        try:
            self.user = self.get_my_user_info_api()
        except VesslApiException as e:
            self.api_client.set_default_header("Authorization", "")
            raise InvalidTokenError("Expired. Please renew your access token.")

        return access_token

    def set_organization(self, organization_name: str = None, credentials_file: str = None) -> str:
        organization_name = self.find_organization_name(organization_name, credentials_file)
        if organization_name:
            try:
                self.organization = self.organization_read_api(organization_name=organization_name)
                return self.organization.name
            except VesslApiException as e:
                raise InvalidOrganizationError(f"Invalid organization {organization_name}.")

        organizations = self.organization_list_api().organizations
        if len(organizations) == 0:
            raise InvalidOrganizationError(
                "No organizations. Create one using vessl.create_organization.",  # TODO: update message
            )
        self.organization = organizations[0]
        return self.organization.name

    def set_project(self, project_name: str = None, credentials_file: str = None) -> str:
        if self.organization is None:
            raise InvalidOrganizationError("Please specify an organization first.")

        projects = self.project_list_api(organization_name=self.organization.name).results
        if len(projects) == 0:
            raise InvalidProjectError(
                "No projects. Create one using vessl.create_project.",  # TODO: update message
            )

        if len(projects) == 1 and project_name is None:
            project_name = projects[0].name
        elif len(projects) > 1:
            project_name = self.find_project_name(project_name, credentials_file)

        for project in projects:
            if project.name == project_name:
                self.project = project
                break
        else:
            raise InvalidProjectError(
                f'Invalid project "{project_name}".',
            )

        return project_name

    def configure_access_token(
        self,
        access_token: str = None,
        credentials_file: str = None,
        force_update: bool = False,
    ) -> None:
        access_token = self.set_access_token(access_token, credentials_file, force_update)
        self.config_loader.access_token = access_token

    def configure_default_organization(
        self, organization_name: str = None, credentials_file: str = None
    ) -> None:
        organization_name = self.set_organization(organization_name, credentials_file)
        self.config_loader.default_organization = organization_name

    def configure_default_project(
        self, project_name: str = None, credentials_file: str = None
    ) -> None:
        project_name = self.set_project(project_name, credentials_file)
        self.config_loader.default_project = project_name

    def set_configurations(
        self,
        access_token: str = None,
        organization_name: str = None,
        project_name: str = None,
        credentials_file: str = None,
    ):
        """Used in vessl.init()"""
        self.set_access_token(access_token, credentials_file)
        self.set_organization(organization_name, credentials_file)
        self.set_project(project_name, credentials_file)



import os
from typing import List, Tuple

import boto3
import botocore.client
from boto3.s3.transfer import S3Transfer
from botocore.credentials import RefreshableCredentials
from botocore.session import Session as BotocoreSession

from vessl import vessl_api
from vessl.cli._util import sizeof_fmt
from vessl.util.common import remove_prefix
from vessl.util.downloader import Downloader
from vessl.util.exception import (
    InvalidParamsError,
    InvalidVolumeFileError,
    VesslApiException,
)


class FileTransmission:
    def __init__(self, source_path, source_abs_path, dest_abs_path, size):
        self.source_path = source_path
        self.source_abs_path = source_abs_path
        self.dest_abs_path = dest_abs_path
        self.size = size


class VolumeFileTransfer:
    def __init__(self, volume_id: int):
        self.volume = vessl_api.volume_read_api(volume_id=volume_id)
        self.provider = None
        self.region_name = None
        self.bucket_name = None
        self.prefix = None
        self._update_federation_credentials()

    def download(self, source_path, dest_path):
        if self.provider == "gs":
            result = vessl_api.volume_file_list_api(
                volume_id=self.volume.id,
                recursive=True,
                path=source_path,
                need_download_url=True,
            ).results
            files = sorted(result, key=lambda x: x.path)
            Downloader.download(source_path, dest_path, *files, quiet=True)
            return

        s3_client = self._get_s3_client()
        (
            file_transmissions,
            total_size,
        ) = self._get_download_file_transmissions_and_total_size(source_path, dest_path)
        total_file_count = len(file_transmissions)
        if total_file_count == 0:
            print("No files to download.")
            return

        formatted_total_size = sizeof_fmt(total_size)
        print(f"Downloading {total_file_count} file(s) ({formatted_total_size})...")

        succeeded_count = 0
        succeeded_size = 0
        for file_transmission in file_transmissions:
            dirname = os.path.dirname(file_transmission.dest_abs_path)
            if dirname:
                os.makedirs(dirname, exist_ok=True)
            try:
                s3_client.download_file(
                    self.bucket_name,
                    file_transmission.source_abs_path,
                    file_transmission.dest_abs_path,
                    Callback=self._create_callback(succeeded_size, total_size),
                )
                succeeded_count += 1
                succeeded_size += file_transmission.size
            except BaseException as e:
                print(f"Failed to download {file_transmission.source_path}.")
                raise e

        print(f"Total {succeeded_count} file(s) downloaded.")

    def upload(self, source_path, dest_path):
        if self.volume.is_read_only:
            print("Cannot upload to read-only volume.")
            return

        s3_client = self._get_s3_client()
        (
            file_transmissions,
            total_size,
        ) = self._get_upload_file_transmissions_and_total_size(source_path, dest_path)
        total_file_count = len(file_transmissions)
        if total_file_count == 0:
            print("No files to upload.")
            return

        formatted_total_size = sizeof_fmt(total_size)
        print(f"Uploading {total_file_count} file(s) ({formatted_total_size})...")

        succeeded_count = 0
        succeeded_size = 0
        succeeded_files = []
        for file_transmission in file_transmissions:
            try:
                s3_client.upload_file(
                    file_transmission.source_abs_path,
                    self.bucket_name,
                    file_transmission.dest_abs_path,
                    Callback=self._create_callback(succeeded_size, total_size),
                )
                succeeded_count += 1
                succeeded_size += file_transmission.size
                succeeded_files.append({"path": os.path.basename(file_transmission.dest_abs_path)})
            except BaseException as e:
                print(f"Failed to upload {file_transmission.source_path}.")
                raise e

        print(f"Total {succeeded_count} file(s) uploaded.")
        return succeeded_files

    def copy(self, source_path, dest_path):
        # TODO
        return

    def remove(self, path):
        # TODO
        return

    def _create_callback(self, current_size, total_size):
        if total_size < 100 * 1024 * 1024:
            return None

        total_transmitted = current_size
        last_percent = int(total_transmitted / total_size * 100)
        formatted_size = sizeof_fmt(total_size)
        interval = max(int(31 * 1024 * 1024 / total_size * 100), 3)

        def callback(transmitted_bytes):
            nonlocal total_transmitted, last_percent
            total_transmitted += transmitted_bytes
            percent = int(total_transmitted / total_size * 100)
            if percent - interval >= last_percent or percent == 100:
                print(f"{sizeof_fmt(total_transmitted)}/{formatted_size} ({percent}%) completed.")
                last_percent = percent

        return callback

    def _update_federation_credentials(self):
        federation_credentials = vessl_api.volume_federate_api(volume_id=self.volume.id)
        self.region_name = federation_credentials.region
        self.bucket_name = federation_credentials.bucket
        self.prefix = federation_credentials.prefix
        self.provider = federation_credentials.token.provider

    def _get_s3_client(self) -> S3Transfer:
        refreshable_credentials = RefreshableCredentials.create_from_metadata(
            metadata=self.__get_s3_session_credentials(),
            refresh_using=self.__get_s3_session_credentials,
            method="sts-assume-role",
        )

        botocore_session = BotocoreSession()
        botocore_session._credentials = refreshable_credentials
        botocore_session.set_config_variable("region", self.region_name)
        session = boto3.session.Session(botocore_session=botocore_session)
        config = botocore.client.Config(connect_timeout=120, read_timeout=120)
        return session.client("s3", config=config)

    def _get_download_file_transmissions_and_total_size(
        self, source_path: str, dest_path: str
    ) -> Tuple[List[FileTransmission], int]:
        source_abs_path = f"{self.prefix}/{source_path}" if self.prefix else source_path
        dest_abs_path = os.path.abspath(dest_path)
        try:
            source_file = vessl_api.volume_file_read_api(volume_id=self.volume.id, path=source_path)
        except VesslApiException:
            files = vessl_api.volume_file_list_api(
                volume_id=self.volume.id,
                recursive=True,
                path=source_path,
                need_download_url=False,
            ).results
            # source path is a directory (if not, raise an error)

            if os.path.exists(dest_abs_path):
                if os.path.isfile(dest_abs_path):
                    raise InvalidVolumeFileError(
                        f"Destination path is not a directory: {dest_path}."
                    )
            else:
                os.makedirs(dest_abs_path)

            file_transmissions = []
            total_size = 0
            for file in files:
                if file.is_dir:
                    continue
                file_transmissions.append(
                    FileTransmission(
                        file.path,
                        os.path.join(self.prefix, file.path.lstrip("/")),
                        os.path.join(
                            dest_abs_path,
                            remove_prefix(file.path, source_path).lstrip("/"),
                        ),
                        file.size,
                    )
                )
                total_size += file.size
            return file_transmissions, total_size

        # source path is a file
        file_name = os.path.basename(source_file.path)
        if os.path.isdir(dest_abs_path):
            dest_abs_path = os.path.join(dest_abs_path, file_name)
            if os.path.isdir(dest_abs_path):
                # Case where `source_path` is "a.txt", `dest_abs_path` is "dir/", and
                # "dir/a.txt/" exists as a directory
                raise InvalidParamsError(
                    f"Cannot overwrite directory {dest_abs_path} with non-directory {file_name}."
                )

        return [
            FileTransmission(
                source_path,
                source_abs_path,
                dest_abs_path,
                source_file.size,
            )
        ], source_file.size

    def _get_upload_file_transmissions_and_total_size(
        self, source_path: str, dest_path: str
    ) -> Tuple[List[FileTransmission], int]:
        source_path = os.path.relpath(source_path)
        if not os.path.exists(source_path):
            raise InvalidParamsError(f"{source_path} does not exist.")

        source_file_name = os.path.basename(source_path)
        if os.path.isfile(source_path):
            try:
                vessl_api.volume_file_read_api(volume_id=self.volume.id, path=dest_path)
                dest_abs_path = dest_path
            except VesslApiException as e:
                if e.code == "NotAFile":
                    dest_abs_path = f"{dest_path}/{source_file_name}"
                elif e.code == "NotFound":
                    dest_abs_path = dest_path
                else:
                    raise

            if self.prefix:
                dest_abs_path = os.path.join(self.prefix, dest_abs_path.lstrip("/"))

            source_abs_path = os.path.abspath(source_path)
            file_size = os.path.getsize(source_abs_path)
            return [
                FileTransmission(
                    source_path,
                    source_abs_path,
                    dest_abs_path,
                    file_size,
                )
            ], file_size
        else:
            try:
                vessl_api.volume_file_read_api(volume_id=self.volume.id, path=dest_path)
            except VesslApiException as e:
                if e.code not in ("NotAFile", "NotFound"):
                    raise

                result = []
                total_size = 0
                for root_path, _, file_names in os.walk(source_path):
                    for file_name in file_names:
                        source_abs_path = os.path.join(root_path, file_name)
                        dest_abs_path = os.path.join(
                            dest_path,
                            remove_prefix(source_abs_path, source_path).lstrip("/"),
                        )
                        if self.prefix:
                            dest_abs_path = os.path.join(self.prefix, dest_abs_path.lstrip("/"))
                        file_size = os.path.getsize(source_abs_path)
                        result.append(
                            FileTransmission(
                                source_path,
                                source_abs_path,
                                dest_abs_path,
                                file_size,
                            )
                        )
                        total_size += file_size

                return result, total_size

            raise InvalidVolumeFileError(f"Destination path is not a directory: {dest_path}.")

    def __get_s3_session_credentials(self):
        federation_credentials = vessl_api.volume_federate_api(volume_id=self.volume.id)
        creds = federation_credentials.token.s3
        return {
            "access_key": creds.access_key_id,
            "secret_key": creds.secret_access_key,
            "token": creds.session_token,
            "expiry_time": creds.expiration.isoformat(),
        }



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from openapi_client.configuration import Configuration


class OrmProjectDashboardEdges(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'chart_sections': 'list[OrmProjectDashboardChartSection]',
        'created_by': 'OrmUser',
        'field_configs': 'list[OrmProjectDashboardExperimentFieldConfig]',
        'filters': 'list[OrmProjectDashboardExperimentFilter]',
        'hide_plots_experiments': 'list[OrmExperiment]',
        'last_updated_by': 'OrmUser',
        'project': 'OrmProject',
        'sorts': 'list[OrmProjectDashboardExperimentSort]',
        'starred_users': 'list[OrmUser]'
    }

    attribute_map = {
        'chart_sections': 'chart_sections',
        'created_by': 'created_by',
        'field_configs': 'field_configs',
        'filters': 'filters',
        'hide_plots_experiments': 'hide_plots_experiments',
        'last_updated_by': 'last_updated_by',
        'project': 'project',
        'sorts': 'sorts',
        'starred_users': 'starred_users'
    }

    def __init__(self, chart_sections=None, created_by=None, field_configs=None, filters=None, hide_plots_experiments=None, last_updated_by=None, project=None, sorts=None, starred_users=None, local_vars_configuration=None):  # noqa: E501
        """OrmProjectDashboardEdges - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._chart_sections = None
        self._created_by = None
        self._field_configs = None
        self._filters = None
        self._hide_plots_experiments = None
        self._last_updated_by = None
        self._project = None
        self._sorts = None
        self._starred_users = None
        self.discriminator = None

        if chart_sections is not None:
            self.chart_sections = chart_sections
        if created_by is not None:
            self.created_by = created_by
        if field_configs is not None:
            self.field_configs = field_configs
        if filters is not None:
            self.filters = filters
        if hide_plots_experiments is not None:
            self.hide_plots_experiments = hide_plots_experiments
        if last_updated_by is not None:
            self.last_updated_by = last_updated_by
        if project is not None:
            self.project = project
        if sorts is not None:
            self.sorts = sorts
        if starred_users is not None:
            self.starred_users = starred_users

    @property
    def chart_sections(self):
        """Gets the chart_sections of this OrmProjectDashboardEdges.  # noqa: E501


        :return: The chart_sections of this OrmProjectDashboardEdges.  # noqa: E501
        :rtype: list[OrmProjectDashboardChartSection]
        """
        return self._chart_sections

    @chart_sections.setter
    def chart_sections(self, chart_sections):
        """Sets the chart_sections of this OrmProjectDashboardEdges.


        :param chart_sections: The chart_sections of this OrmProjectDashboardEdges.  # noqa: E501
        :type chart_sections: list[OrmProjectDashboardChartSection]
        """

        self._chart_sections = chart_sections

    @property
    def created_by(self):
        """Gets the created_by of this OrmProjectDashboardEdges.  # noqa: E501


        :return: The created_by of this OrmProjectDashboardEdges.  # noqa: E501
        :rtype: OrmUser
        """
        return self._created_by

    @created_by.setter
    def created_by(self, created_by):
        """Sets the created_by of this OrmProjectDashboardEdges.


        :param created_by: The created_by of this OrmProjectDashboardEdges.  # noqa: E501
        :type created_by: OrmUser
        """

        self._created_by = created_by

    @property
    def field_configs(self):
        """Gets the field_configs of this OrmProjectDashboardEdges.  # noqa: E501


        :return: The field_configs of this OrmProjectDashboardEdges.  # noqa: E501
        :rtype: list[OrmProjectDashboardExperimentFieldConfig]
        """
        return self._field_configs

    @field_configs.setter
    def field_configs(self, field_configs):
        """Sets the field_configs of this OrmProjectDashboardEdges.


        :param field_configs: The field_configs of this OrmProjectDashboardEdges.  # noqa: E501
        :type field_configs: list[OrmProjectDashboardExperimentFieldConfig]
        """

        self._field_configs = field_configs

    @property
    def filters(self):
        """Gets the filters of this OrmProjectDashboardEdges.  # noqa: E501


        :return: The filters of this OrmProjectDashboardEdges.  # noqa: E501
        :rtype: list[OrmProjectDashboardExperimentFilter]
        """
        return self._filters

    @filters.setter
    def filters(self, filters):
        """Sets the filters of this OrmProjectDashboardEdges.


        :param filters: The filters of this OrmProjectDashboardEdges.  # noqa: E501
        :type filters: list[OrmProjectDashboardExperimentFilter]
        """

        self._filters = filters

    @property
    def hide_plots_experiments(self):
        """Gets the hide_plots_experiments of this OrmProjectDashboardEdges.  # noqa: E501


        :return: The hide_plots_experiments of this OrmProjectDashboardEdges.  # noqa: E501
        :rtype: list[OrmExperiment]
        """
        return self._hide_plots_experiments

    @hide_plots_experiments.setter
    def hide_plots_experiments(self, hide_plots_experiments):
        """Sets the hide_plots_experiments of this OrmProjectDashboardEdges.


        :param hide_plots_experiments: The hide_plots_experiments of this OrmProjectDashboardEdges.  # noqa: E501
        :type hide_plots_experiments: list[OrmExperiment]
        """

        self._hide_plots_experiments = hide_plots_experiments

    @property
    def last_updated_by(self):
        """Gets the last_updated_by of this OrmProjectDashboardEdges.  # noqa: E501


        :return: The last_updated_by of this OrmProjectDashboardEdges.  # noqa: E501
        :rtype: OrmUser
        """
        return self._last_updated_by

    @last_updated_by.setter
    def last_updated_by(self, last_updated_by):
        """Sets the last_updated_by of this OrmProjectDashboardEdges.


        :param last_updated_by: The last_updated_by of this OrmProjectDashboardEdges.  # noqa: E501
        :type last_updated_by: OrmUser
        """

        self._last_updated_by = last_updated_by

    @property
    def project(self):
        """Gets the project of this OrmProjectDashboardEdges.  # noqa: E501


        :return: The project of this OrmProjectDashboardEdges.  # noqa: E501
        :rtype: OrmProject
        """
        return self._project

    @project.setter
    def project(self, project):
        """Sets the project of this OrmProjectDashboardEdges.


        :param project: The project of this OrmProjectDashboardEdges.  # noqa: E501
        :type project: OrmProject
        """

        self._project = project

    @property
    def sorts(self):
        """Gets the sorts of this OrmProjectDashboardEdges.  # noqa: E501


        :return: The sorts of this OrmProjectDashboardEdges.  # noqa: E501
        :rtype: list[OrmProjectDashboardExperimentSort]
        """
        return self._sorts

    @sorts.setter
    def sorts(self, sorts):
        """Sets the sorts of this OrmProjectDashboardEdges.


        :param sorts: The sorts of this OrmProjectDashboardEdges.  # noqa: E501
        :type sorts: list[OrmProjectDashboardExperimentSort]
        """

        self._sorts = sorts

    @property
    def starred_users(self):
        """Gets the starred_users of this OrmProjectDashboardEdges.  # noqa: E501


        :return: The starred_users of this OrmProjectDashboardEdges.  # noqa: E501
        :rtype: list[OrmUser]
        """
        return self._starred_users

    @starred_users.setter
    def starred_users(self, starred_users):
        """Sets the starred_users of this OrmProjectDashboardEdges.


        :param starred_users: The starred_users of this OrmProjectDashboardEdges.  # noqa: E501
        :type starred_users: list[OrmUser]
        """

        self._starred_users = starred_users

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, OrmProjectDashboardEdges):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, OrmProjectDashboardEdges):
            return True

        return self.to_dict() != other.to_dict()



import os
import zipfile

from vessl.util import logger


class Zipper(zipfile.ZipFile):
    def __init__(self, file, mode):
        super(Zipper, self).__init__(file, mode)

    def zip(self, path):
        if not os.path.isabs(path):
            path = os.path.isabs(path)
        if os.path.isdir(path):
            self.zipdir(path)
        elif os.path.isfile(path):
            self.zipfile(path)

    def zipfile(self, path):
        self.write(path, os.path.basename(path))

    def zipdir(self, path, include_dotfiles=False):
        for dir_full_path, dirs, files in os.walk(path):
            if os.path.basename(dir_full_path) == ".vessl":
                continue

            if not include_dotfiles:
                files = [f for f in files if not f.startswith(".")]

            for file in files:
                filename = os.path.join(dir_full_path, file)
                try:
                    os.stat(filename)
                except FileNotFoundError:
                    continue
                self.write(filename, os.path.relpath(os.path.join(dir_full_path, file), path))
                logger.debug(f"Compressed {filename}.")

    def size(self):
        return os.path.getsize(self.filename)

    def remove(self):
        os.remove(self.filename)



import os
import time
from typing import List

import click

from vessl.cli._base import VesslGroup, vessl_argument, vessl_option
from vessl.cli._util import (
    Endpoint,
    generic_prompter,
    print_data,
    print_logs,
    print_table,
    print_volume_files,
    prompt_choices,
    sizeof_fmt,
    truncate_datetime,
)
from vessl.cli.kernel_cluster import cluster_option
from vessl.cli.kernel_resource_spec import resource_option
from vessl.cli.organization import organization_name_option
from vessl.cli.project import project_name_option
from vessl.experiment import (
    create_experiment,
    delete_experiment,
    download_experiment_output_files,
    list_experiment_logs,
    list_experiment_output_files,
    list_experiments,
    list_github_code_refs,
    read_experiment,
    terminate_experiment,
)
from vessl.kernel_cluster import list_cluster_nodes
from vessl.kernel_image import list_kernel_images
from vessl.util import logger
from vessl.util.common import safe_cast
from vessl.util.constant import (
    EXPERIMENT_WORKING_DIR,
    FRAMEWORK_TYPE_PYTORCH,
    FRAMEWORK_TYPE_TENSORFLOW,
    FRAMEWORK_TYPES,
    MOUNT_PATH_OUTPUT,
    PROCESSOR_TYPE_GPU,
    PROCESSOR_TYPES,
)
from vessl.util.git_remote import clone_codes


def experiment_number_prompter(
    ctx: click.Context,
    param: click.Parameter,
    value: str,
) -> str:
    experiments = list_experiments()
    return prompt_choices("Experiment", [(f"#{x.number}", x.number) for x in reversed(experiments)])


def local_experiment_number_prompter(
    ctx: click.Context,
    param: click.Parameter,
    value: str,
) -> str:
    experiments = [e for e in list_experiments() if e.is_local]
    return prompt_choices("Experiment", [(f"#{x.number}", x.number) for x in reversed(experiments)])


def processor_type_prompter(ctx: click.Context, param: click.Parameter, value: str) -> str:
    cluster = ctx.obj.get("cluster")
    if cluster is None:
        raise click.BadOptionUsage(
            option_name="--cluster",
            message="Cluster (`--cluster`) must be specified before processor type (`--processor-type`).",
        )

    if ctx.obj.get("resource") is None:
        processor_type = prompt_choices("Processor Type", PROCESSOR_TYPES)
        ctx.obj["processor_type"] = processor_type
        return processor_type


def processor_type_callback(ctx: click.Context, param: click.Parameter, value: str):
    if value:
        ctx.obj["processor_type"] = value
    return value


def cpu_limit_prompter(ctx: click.Context, param: click.Parameter, value: float) -> float:
    cluster = ctx.obj.get("cluster")
    if cluster is None:
        raise click.BadOptionUsage(
            option_name="--cluster",
            message="Cluster (`--cluster`) must be specified before CPU limit (`--cpu-limit`).",
        )

    if ctx.obj.get("resource") is None:
        return click.prompt("CPUs (in vCPU)", type=click.FLOAT)


def memory_limit_prompter(ctx: click.Context, param: click.Parameter, value: float) -> float:
    cluster = ctx.obj.get("cluster")
    if cluster is None:
        raise click.BadOptionUsage(
            option_name="--cluster",
            message="Cluster (`--cluster`) must be specified before memory limit (`--memory-limit`).",
        )

    if ctx.obj.get("resource") is None:
        return click.prompt("Memory (e.g. 4Gi)", type=click.STRING)


def gpu_type_prompter(ctx: click.Context, param: click.Parameter, value: str) -> str:
    cluster = ctx.obj.get("cluster")
    if cluster is None:
        raise click.BadOptionUsage(
            option_name="--cluster",
            message="Cluster (`--cluster`) must be specified before GPU type (`--gpu-type`).",
        )

    if ctx.obj.get("resource") is None:
        processor_type = ctx.obj.get("processor_type")
        if processor_type is None:
            raise click.UsageError(
                message="Processor type must be specified before GPU type (`--gpu-type`).",
            )
        if processor_type == PROCESSOR_TYPE_GPU:
            nodes = list_cluster_nodes(cluster.id)
            return prompt_choices("GPU Type", [x.gpu_product_name for x in nodes])


def gpu_limit_prompter(ctx: click.Context, param: click.Parameter, value: float) -> float:
    cluster = ctx.obj.get("cluster")
    if cluster is None:
        raise click.BadOptionUsage(
            option_name="--cluster",
            message="Cluster (`--cluster`) must be specified before GPU limit (`--gpu-limit`).",
        )

    if ctx.obj.get("resource") is None:
        processor_type = ctx.obj.get("processor_type")
        if processor_type is None:
            raise click.UsageError(
                message="Processor type must be specified before GPU limit (`--gpu-limit`).",
            )

        if processor_type == PROCESSOR_TYPE_GPU:
            return click.prompt("GPUs (in vGPU)", type=click.FLOAT)


def image_url_prompter(ctx: click.Context, param: click.Parameter, value: str) -> str:
    processor_type = ctx.obj.get("processor_type")
    if processor_type is None:
        raise click.UsageError(
            message="Processor type must be specified before image URL (`--image-url`).",
        )

    images = list_kernel_images()
    images = [x for x in images if x.processor_type == processor_type]

    # Pretty print image name+URL choices
    name_maxlen = len(max(images, key=lambda x: len(x.name)).name)

    return prompt_choices(
        "Image",
        [
            (
                f"{x.name}" + " " * (name_maxlen - len(x.name) + 2) + f"(URL: {x.image_url})",
                x.image_url,
            )
            for x in images
        ],
    )


def upload_local_file_callback(
    ctx: click.Context, param: click.Parameter, value: List[str]
) -> List[str]:
    if value:
        for dir_spec in value:
            if ":" in dir_spec:
                local_path, _ = dir_spec.split(":")
            else:
                local_path = dir_spec

            if not local_path:
                raise click.BadOptionUsage(
                    option_name="--upload-local-file",
                    message=f"Invalid path {local_path}.",
                )

            if not os.path.exists(local_path):
                raise click.BadOptionUsage(
                    option_name="--upload-local-file",
                    message=f"{local_path} does not exist.",
                )

            if os.path.isdir(local_path):
                from pathlib import Path

                root_directory = Path(local_path)
                directory_size = sum(
                    f.stat().st_size for f in root_directory.glob("**/*") if f.is_file()
                )
                logger.debug(f"Current directory size: {sizeof_fmt(directory_size)}")
                if directory_size >= 3 * 1024 * 1024 * 1024:
                    click.confirm(
                        f"Try to upload [{os.path.abspath(local_path)}] "
                        f"but it's too large ({sizeof_fmt(directory_size)}). "
                        f"Process anyway?",
                        abort=True,
                    )

    return value


def worker_count_callback(ctx: click.Context, param: click.Parameter, value: int) -> int:
    if value is None:
        value = 1

    if value < 1:
        raise click.BadOptionUsage(
            option_name="--worker-count",
            message="num nodes (`--num-nodes`) must be a positive integer.",
        )

    ctx.obj["worker_count"] = value
    return value


def framework_type_prompter(ctx: click.Context, param: click.Parameter, value: str) -> str:
    worker_count = ctx.obj.get("worker_count")
    if worker_count == 1:
        return ""

    framework_type = prompt_choices("Processor Type", FRAMEWORK_TYPES)
    if framework_type == FRAMEWORK_TYPE_TENSORFLOW:
        raise click.BadOptionUsage(
            option_name="--framework-type",
            message="Only PyTorch distributed experiment is supported currently.",
        )
    return framework_type


@click.command(name="experiment", cls=VesslGroup)
def cli():
    pass


@cli.vessl_command()
@vessl_argument(
    "number",
    type=click.INT,
    required=True,
    prompter=experiment_number_prompter,
)
@organization_name_option
@project_name_option
def read(number: int):
    experiment = read_experiment(experiment_number=number)

    distributed_spec = "None"
    if experiment.is_distributed:
        if experiment.distributed_spec.framework_type == FRAMEWORK_TYPE_PYTORCH:
            distributed_spec = {
                "Framework Type": experiment.distributed_spec.framework_type,
                "PyTorch Spec": {
                    "Worker count": experiment.distributed_spec.pytorch_spec.worker_replicas,
                },
            }
        elif experiment.distributed_spec.framework_type == FRAMEWORK_TYPE_TENSORFLOW:
            distributed_spec = {
                "Framework Type": experiment.distributed_spec.framework_type,
                "TensorFlow Spec": experiment.distributed_spec.tensorflow_spec,
            }

    kernel_image = "None"
    if experiment.kernel_image:
        kernel_image = {
            "Name": experiment.kernel_image.name,
            "URL": experiment.kernel_image.image_url,
        }

    resource_spec = "None"
    if experiment.kernel_resource_spec:
        resource_spec = {
            "Name": experiment.kernel_resource_spec.name,
            "CPU Type": experiment.kernel_resource_spec.cpu_type,
            "CPU Limit": experiment.kernel_resource_spec.cpu_limit,
            "Memory Limit": experiment.kernel_resource_spec.memory_limit,
            "GPU Type": experiment.kernel_resource_spec.gpu_type,
            "GPU Limit": experiment.kernel_resource_spec.gpu_limit,
        }

    metrics_summary = "None"
    if experiment.metrics_summary.latest:
        metrics_keys = experiment.metrics_summary.latest.keys()
        metrics_summary = {}
        for key in metrics_keys:
            metrics_summary[key] = experiment.metrics_summary.latest[key].value

    print_data(
        {
            "ID": experiment.id,
            "Number": experiment.number,
            "Distributed": experiment.is_distributed,
            "Distributed spec": distributed_spec,
            "Local": experiment.is_local,
            "Status": experiment.status,
            "Created": truncate_datetime(experiment.created_dt),
            "Message": experiment.message,
            "Kernel image": kernel_image,
            "Resource spec": resource_spec,
            "Start command": experiment.start_command,
            "Metrics summary": metrics_summary,
        }
    )
    print(
        f"For more info: {Endpoint.experiment.format(experiment.organization.name, experiment.project.name, experiment.number)}"
    )


@cli.vessl_command()
@organization_name_option
@project_name_option
def list():
    experiments = list_experiments()
    print_table(
        experiments,
        ["ID", "Number", "Distributed", "Status", "Created", "Message"],
        lambda x: [
            x.id,
            x.number,
            x.is_distributed,
            x.status,
            truncate_datetime(x.created_dt),
            x.message,
        ],
    )


command_option = vessl_option(
    "-x",
    "--command",
    type=click.STRING,
    required=True,
    prompter=generic_prompter("Start command"),
    help="Start command to execute in experiment container.",
)
processor_type_option = vessl_option(
    "--processor-type",
    type=click.Choice(("CPU", "GPU")),
    prompter=processor_type_prompter,
    callback=processor_type_callback,
    help="CPU or GPU (for custom resource only).",
)
cpu_limit_option = vessl_option(
    "--cpu-limit",
    type=click.FLOAT,
    prompter=cpu_limit_prompter,
    help="Number of vCPUs (for custom resource only).",
)
memory_limit_option = vessl_option(
    "--memory-limit",
    type=click.STRING,
    prompter=memory_limit_prompter,
    help="Memory limit (e.g. 4Gi) (for custom resource only).",
)
gpu_type_option = vessl_option(
    "--gpu-type",
    type=click.STRING,
    prompter=gpu_type_prompter,
    help="GPU type such as Tesla-K80 (for custom resource only).",
)
gpu_limit_option = vessl_option(
    "--gpu-limit",
    type=click.INT,
    prompter=gpu_limit_prompter,
    help="Number of GPU cores (for custom resource only).",
)
image_url_option = vessl_option(
    "-i",
    "--image-url",
    type=click.STRING,
    prompter=image_url_prompter,
    help="Kernel docker image URL",
)
message_option = click.option("-m", "--message", type=click.STRING)
termination_protection_option = click.option("--termination-protection", is_flag=True)
hyperparameter_option = click.option(
    "-h",
    "--hyperparameter",
    type=click.STRING,
    multiple=True,
    help="Hyperparameters. Format: [key]=[value], ex. `--hyperparameter lr=0.01`.",
)
dataset_option = click.option(
    "--dataset",
    type=click.STRING,
    multiple=True,
    help="Dataset mounts. Format: [mount_path]:[dataset_name]@[optional_dataset_version], ex. `--dataset /input:mnist@3bcd5f`.",
)
model_option = click.option(
    "--model",
    type=click.STRING,
    multiple=True,
    help="Model mounts. Format: [mount_path]:[model_repository_name]/[model_number], ex. `--model /input:pytorch-model/1` .",
)
git_ref_option = click.option(
    "--git-ref",
    type=click.STRING,
    multiple=True,
    help=f"Git repository mounts. Format: [mount_path]:github/[organization]/[repository]/[optional_commit], ex. `--git-ref {EXPERIMENT_WORKING_DIR}examples:github/vessl-ai/examples/3cd23dd`.",
)
git_diff_option = click.option(
    "--git-diff",
    type=click.STRING,
    help="Git diff file mounts. Format: [mount_path]:[volume_file_path]. This option is used only for reproducing existing experiments.",
)
upload_local_file_option = vessl_option(
    "--upload-local-file",
    type=click.STRING,
    multiple=True,
    callback=upload_local_file_callback,
    help="Upload local file. Format: [local_path] or [local_path]:[remote_path].",
)
upload_local_git_diff_option = vessl_option(
    "--upload-local-git-diff",
    is_flag=True,
    help="Upload local git commit hash and diff.",
)
archive_file_option = click.option(
    "--archive-file",
    type=click.STRING,
    help="Local archive file mounts. Format: [mount_path]:[archive_file_path]]. This option is used only for reproducing existing experiments.",
)
root_volume_size_option = click.option("--root-volume-size", type=click.STRING)
working_dir_option = click.option(
    "--working-dir", type=click.STRING, help=f"Defaults to `{EXPERIMENT_WORKING_DIR}`."
)
output_dir_option = click.option(
    "--output-dir",
    type=click.STRING,
    default=MOUNT_PATH_OUTPUT,
    help="Directory to store experiment output files. Defaults to `/output/`.",
)
worker_count_option = vessl_option(
    "--worker-count",
    type=click.INT,
    callback=worker_count_callback,
    help="The number of nodes to run an experiment. Defaults to 1.",
)
framework_type_option = vessl_option(
    "--framework-type",
    type=click.STRING,
    prompter=framework_type_prompter,
    help="Framework type option. Defaults to `pytorch`.",
)


@cli.vessl_command()
@click.pass_context
@organization_name_option
@project_name_option
@cluster_option
@vessl_option(
    "--node",
    type=click.STRING,
    multiple=True,
    help="Cluster nodes. Defaults to all nodes in cluster.",
)
@resource_option
@processor_type_option
@cpu_limit_option
@memory_limit_option
@gpu_type_option
@gpu_limit_option
@image_url_option
@message_option
@termination_protection_option
@hyperparameter_option
@dataset_option
@model_option
@git_ref_option
@git_diff_option
@upload_local_file_option
@upload_local_git_diff_option
@archive_file_option
@root_volume_size_option
@working_dir_option
@output_dir_option
@command_option
@worker_count_option
@framework_type_option
def create(
    ctx,
    cluster: str,
    node: List[str],
    command: str,
    resource: str,
    processor_type: str,
    cpu_limit: float,
    memory_limit: str,
    gpu_type: str,
    gpu_limit: int,
    image_url: str,
    message: str,
    termination_protection: bool,
    hyperparameter: List[str],
    dataset: List[str],
    model: List[str],
    git_ref: List[str],
    git_diff: str,
    upload_local_file: List[str],
    upload_local_git_diff: bool,
    archive_file: str,
    root_volume_size: str,
    working_dir: str,
    output_dir: str,
    worker_count: int,
    framework_type: str,
):
    experiment = create_experiment(
        cluster_name=cluster,
        cluster_node_names=node,
        start_command=command,
        kernel_resource_spec_name=resource,
        processor_type=processor_type,
        cpu_limit=cpu_limit,
        memory_limit=memory_limit,
        gpu_type=gpu_type,
        gpu_limit=gpu_limit,
        kernel_image_url=image_url,
        message=message,
        termination_protection=termination_protection,
        hyperparameters=hyperparameter,
        dataset_mounts=dataset,
        model_mounts=model,
        git_ref_mounts=git_ref,
        git_diff_mount=git_diff,
        local_files=upload_local_file,
        upload_local_git_diff=upload_local_git_diff,
        archive_file_mount=archive_file,
        root_volume_size=root_volume_size,
        working_dir=working_dir,
        output_dir=output_dir,
        worker_count=worker_count,
        framework_type=framework_type,
    )
    ctx.obj["experiment_number"] = experiment.number
    print(
        f"Created '{experiment.number}'.\n"
        f"For more info: {Endpoint.experiment.format(experiment.organization.name, experiment.project.name, experiment.number)}"
    )


worker_number_option = vessl_option(
    "--worker-number",
    type=click.INT,
    default=0,
    help="Worker number (for distributed experiment only).",
)


@cli.vessl_command()
@vessl_argument(
    "number",
    type=click.INT,
    required=True,
    prompter=experiment_number_prompter,
)
@click.option(
    "--tail",
    type=click.INT,
    default=200,
    help="Number of lines to display (from the end).",
)
@click.option(
    "-f",
    "--follow",
    is_flag=True,
    default=False,
)
@organization_name_option
@project_name_option
@worker_number_option
def logs(
    number: int,
    tail: int,
    follow: bool,
    worker_number: int,
):
    if not follow:
        logs = list_experiment_logs(experiment_number=number, tail=tail, worker_numer=worker_number)
        print_logs(logs)
        print(f"Displayed last {len(logs)} lines of #{number}.")
        return

    after = 0

    experiment_finished_dt = None
    while True:
        if (
            read_experiment(number).status not in ["pending", "running"]
            and experiment_finished_dt is None
        ):
            experiment_finished_dt = time.time()

        if experiment_finished_dt is not None and (time.time() - experiment_finished_dt) > 5:
            break

        logs = list_experiment_logs(
            experiment_number=number,
            before=int(time.time() - 5),
            after=after,
            worker_numer=worker_number,
        )
        print_logs(logs)
        if len(logs) > 0:
            after = logs[-1].timestamp + 0.000001

        time.sleep(3)


@cli.vessl_command()
@vessl_argument(
    "number",
    type=click.INT,
    required=True,
    prompter=experiment_number_prompter,
)
@organization_name_option
@project_name_option
@worker_number_option
def list_output(
    number: int,
    worker_number: int,
):
    files = list_experiment_output_files(
        experiment_number=number,
        need_download_url=False,
        recursive=True,
        worker_number=worker_number,
    )
    print_volume_files(files)


@cli.vessl_command()
@vessl_argument(
    "number",
    type=click.INT,
    required=True,
    prompter=experiment_number_prompter,
)
@click.option(
    "-p",
    "--path",
    type=click.Path(),
    default=os.path.join(os.getcwd(), "output"),
    help="Path to store downloads. Defaults to `./output`.",
)
@worker_number_option
@organization_name_option
@project_name_option
def download_output(
    number: int,
    path: str,
    worker_number: int,
):
    download_experiment_output_files(
        experiment_number=number,
        dest_path=path,
        worker_number=worker_number,
    )
    print(f"Downloaded experiment output to {path}.")


@cli.vessl_command()
@vessl_argument(
    "number",
    type=click.INT,
    required=True,
    prompter=experiment_number_prompter,
)
@organization_name_option
@project_name_option
def terminate(number: int):
    experiment = terminate_experiment(experiment_number=number)
    print(
        f"Terminated '#{experiment.number}'.\n"
        f"For more info: {Endpoint.experiment.format(experiment.organization.name, experiment.project.name, experiment.number)}"
    )


@cli.vessl_command(hidden=True)
def download_code():
    """
    This command is not for users and supposed to run on initContainers of experiment Pods.
    """
    workload_id = safe_cast(os.environ.get("VESSL_WORKLOAD_ID", None), int)
    if workload_id is None:
        print("Failed to download source codes: no workload id.")
        return

    github_code_refs = list_github_code_refs(workload_id)
    if not github_code_refs:
        print("No source code to download.")
        return

    clone_codes(github_code_refs)


@cli.vessl_command()
@vessl_argument(
    "number",
    type=click.INT,
    required=True,
    prompter=experiment_number_prompter,
)
@organization_name_option
@project_name_option
def delete(number: int):
    delete_experiment(experiment_number=number)
    print(f"Deleted '#{number}'.\n")
